# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019-2020 Dinar Batyrshin <batyrshin-dinar@mail.ru>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from CachedMethods import cached_method
from collections import defaultdict
from math import atan2, sin, cos, hypot
from typing import Tuple, TYPE_CHECKING, Union
from uuid import uuid4


if TYPE_CHECKING:
    from chython import ReactionContainer, MoleculeContainer

cpk = tuple('''
 #909090                                                                                         #D9FFFF
 #CC80FF #C2FF00                                         #FFB5B5 #101010 #3050F8 #FF0D0D #90E050 #B3E3F5
 #AB5CF2 #8AFF00                                         #BFA6A6 #F0C8A0 #FF8000 #C6C600 #1FF01F #80D1E3
 #8F40D4 #3DFF00 #E6E6E6 #BFC2C7 #A6A6AB #8A99C7 #9C7AC7 
                 #E06633 #F090A0 #50D050 #C88033 #7D80B0 #C28F8F #668F8F #BD80E3 #FFA100 #A62929 #5CB8D1
 #702EB0 #00FF00 #94FFFF #94E0E0 #73C2C9 #54B5B5 #3B9E9E 
                 #248F8F #0A7D8C #006985 #C0C0C0 #FFD98F #A67573 #668080 #9E63B5 #D47A00 #940094 #429EB0
 #57178F #00C900 #70D4FF 
                 #FFFFC7 #D9FFC7 #C7FFC7 #A3FFC7 #8FFFC7 #61FFC7 #45FFC7
                 #30FFC7 #1FFFC7 #00FF9C #00E675 #00D452 #00BF38 #00AB24 
                         #4DC2FF #4DA6FF #2194D6 #267DAB
                 #266696 #175487 #D0D0E0 #FFD123 #B8B8D0 #A6544D #575961 #9E4FB5 #AB5C00 #754F45 #428296
 #420066 #007D00 #70ABFA
                 #00BAFF #00A1FF #008FFF #0080FF #006BFF #545CF2 #785CE3
                 #8A4FE3 #A136D4 #B31FD4 #B31FBA #B30DA6 #BD0D87 #C70066
                         #CC0059 #D1004F #D90045 #E00038 
                 #E6002E #EB0026 #EB0026 #EB0026 #EB0026 #EB0026 #EB0026 #EB0026 #EB0026 #EB0026 #EB0026
'''.split())
_render_charge = {-4: '4-', -3: '3-', -2: '2-', -1: '-', 1: '+', 2: '2+', 3: '3+', 4: '4+'}
_render_config = {'carbon': False, 'dashes': (.2, .1), 'span_dy': .15, 'mapping': True, 'font_size': .5,
                  'span_size': .35, 'other_size': 0.3, 'monochrome': False, 'bond_color': 'black', 'bond_width': .04,
                  'other_color': 'black', 'bond_radius': .02, 'atom_radius': -.2, 'mapping_size': .25,
                  'atoms_colors': cpk, 'triple_space': .13, 'double_space': .06, 'mapping_color': '#0305A7',
                  'aromatic_space': .14, 'aromatic_dashes': (.15, .05), 'dx_m': .05, 'dy_m': .2,
                  'other_font_style': 'monospace', 'dx_ci': .05, 'dy_ci': 0.2, 'symbols_font_style': 'sans-serif',
                  'mapping_font_style': 'monospace', 'wedge_space': .08}


def _rotate_vector(x1, y1, x2, y2):
    """
    rotate x,y vector over x2-x1, y2-y1 angle
    """
    angle = atan2(y2, x2)
    cos_rad = cos(angle)
    sin_rad = sin(angle)
    return cos_rad * x1 - sin_rad * y1, sin_rad * x1 + cos_rad * y1


def _graph_svg(atoms, bonds, define, masks, uid, viewbox_x, viewbox_y, width, height):
    svg = [f'  <g id="{uid}-molecule">\n    <defs>']
    svg.extend(define)
    if bonds:
        if masks:
            svg.append(f'      <mask id="{uid}-mask">\n'
                       f'        <rect x="{viewbox_x:.2f}" y="{viewbox_y:.2f}" '
                       f'width="{width:.2f}" height="{height:.2f}" fill="white"/>')
            svg.extend(masks)
            svg.append('      </mask>\n    </defs>\n'
                       f'    <g fill="none" stroke="{_render_config["bond_color"]}" '
                       f'stroke-width="{_render_config["bond_width"]:.2f}" mask="url(#{uid}-mask)">')
            if len(bonds) == 1:  # SVG BUG adhoc
                svg.append(f'      <line x1="{viewbox_x:.2f}" y1="{viewbox_y:.2f}" '
                           f'x2="{viewbox_x + width:.2f}" y2="{viewbox_y:.2f}" stroke="none"/>')
        else:
            svg.append(f'    </defs>\n    <g fill="none" stroke="{_render_config["bond_color"]}" '
                       f'stroke-width="{_render_config["bond_width"]:.2f}">')
        svg.extend(bonds)
        svg.append('    </g>')
    else:
        svg.append('    </defs>')

    svg.extend(atoms)
    svg.append('  </g>')
    return svg


def _render_aromatic_bond(n_x, n_y, m_x, m_y, c_x, c_y):
    aromatic_space = _render_config['aromatic_space']
    dash3, dash4 = _render_config['aromatic_dashes']
    # n aligned xy
    mn_x, mn_y, cn_x, cn_y = m_x - n_x, m_y - n_y, c_x - n_x, c_y - n_y

    # nm reoriented xy
    mr_x, mr_y = hypot(mn_x, mn_y), 0
    cr_x, cr_y = _rotate_vector(cn_x, cn_y, mn_x, -mn_y)

    if cr_y and aromatic_space / cr_y < .65:
        if cr_y > 0:
            r_y = aromatic_space
        else:
            r_y = -aromatic_space
            cr_y = -cr_y

        ar_x = aromatic_space * cr_x / cr_y
        br_x = mr_x - aromatic_space * (mr_x - cr_x) / cr_y

        # backward reorienting
        an_x, an_y = _rotate_vector(ar_x, r_y, mn_x, mn_y)
        bn_x, bn_y = _rotate_vector(br_x, r_y, mn_x, mn_y)
        a_x, a_y = n_x + an_x, n_y + an_y
        b_x, b_y = n_x + bn_x, n_y + bn_y

        return f'      <line x1="{a_x:.2f}" y1="{-a_y:.2f}" x2="{b_x:.2f}" y2="{-b_y:.2f}" ' \
               f'stroke-dasharray="{dash3:.2f} {dash4:.2f}"/>'


def depict_settings(*, carbon: bool = False, aam: bool = True, monochrome: bool = False,
                    bond_color: str = 'black', aam_color: str = '#0305A7', atoms_colors: tuple = cpk,
                    bond_width: float = .04, wedge_space: float = .08, dashes: Tuple[float, float] = (.2, .1),
                    aromatic_dashes: Tuple[float, float] = (.15, .05), dx_ci: float = .05, dy_ci: float = .2,
                    dx_m: float = .05, dy_m: float = .2, span_dy: float = .15, double_space: float = .06,
                    triple_space: float = .13, aromatic_space: float = .14, atom_radius: float = .2, bond_radius=.02,
                    font_size: float = .5, other_size: float = .3, span_size: float = .35,  aam_size: float = .25,
                    symbols_font_style: str = 'sans-serif', other_font_style: str = 'monospace',
                    other_color: str = 'black', mapping_font_style: str = 'monospace'):
    """
    Settings for depict of chemical structures

    :param carbon: if True, depict atom C
    :param font_size: font size
    :param aam_size: atom-to-atom mapping font size
    :param span_size: font size for hydrogen count
    :param other_size: isotope, radical, charges, neighbors and hybridization symbols size
    :param bond_width: bond width
    :param bond_color: color of bonds
    :param aam_color: atom-to-atom mapping color
    :param atoms_colors: atom colors where key is atomic number - 1, value is atom color (str)
    :param other_color: color for charges, radicals, isotopes
    :param symbols_font_style: font style for atom symbols
    :param other_font_style: font style for charges, radicals, isotopes, hybridization and neighbors
    :param aam: if True, depict mapping
    :param monochrome: if True, colors of items in molecule not used
    :param dashes: first value is long of visible line, second is long of invisible line
    :param aromatic_space: space between simple and aromatic bonds
    :param triple_space: space between simple and triple bonds
    :param double_space: space between simple and double bonds
    :param aromatic_dashes: first value is long of visible line, second is long of invisible line
    :param atom_radius: radius of atoms spheres in depict3d. if negative is multiplier to covalent radii
    :param bond_radius: radius of bonds spheres in depict3d
    :param dx_ci: x-axis offset relative to the center of the atom symbol for radical, charges, isotope
    :param dy_ci: y-axis offset relative to the center of the atom symbol for radical, charges, isotope
    :param dx_m: x-axis offset relative to the center of the atom symbol for atom-to-atom mapping
    :param dy_m: y-axis offset relative to the center of the atom symbol for atom-to-atom mapping
    :param span_dy: y-axis offset relative to the center of the atom symbol for hydrogen count
    :param mapping_font_style: font style for mapping
    :param wedge_space: wedge bond width
    """
    _render_config['carbon'] = carbon
    _render_config['dashes'] = dashes
    _render_config['span_dy'] = span_dy
    _render_config['mapping'] = aam
    _render_config['font_size'] = font_size
    _render_config['span_size'] = span_size
    _render_config['other_size'] = other_size
    _render_config['monochrome'] = monochrome
    _render_config['bond_color'] = bond_color
    _render_config['bond_width'] = bond_width
    _render_config['other_color'] = other_color
    _render_config['bond_radius'] = bond_radius
    _render_config['atom_radius'] = -atom_radius
    _render_config['mapping_size'] = aam_size
    _render_config['atoms_colors'] = atoms_colors
    _render_config['triple_space'] = triple_space
    _render_config['double_space'] = double_space
    _render_config['mapping_color'] = aam_color
    _render_config['aromatic_space'] = aromatic_space
    _render_config['aromatic_dashes'] = aromatic_dashes
    _render_config['dx_m'], _render_config['dy_m'] = dx_m, dy_m
    _render_config['other_font_style'] = other_font_style
    _render_config['dx_ci'], _render_config['dy_ci'] = dx_ci, dy_ci
    _render_config['symbols_font_style'] = symbols_font_style
    _render_config['mapping_font_style'] = mapping_font_style
    _render_config['wedge_space'] = wedge_space


class DepictMolecule:
    __slots__ = ()

    def depict(self: Union['MoleculeContainer', 'DepictMolecule'], *, width=None, height=None, clean2d: bool = True,
               _embedding=False) -> str:
        """
        Depict molecule in SVG format.

        :param width: set svg width param. by default auto-calculated.
        :param height: set svg height param. by default auto-calculated.
        :param clean2d: calculate coordinates if necessary.
        """
        uid = str(uuid4())
        values = self._plane.values()
        min_x = min(x for x, _ in values)
        max_x = max(x for x, _ in values)
        min_y = min(y for _, y in values)
        max_y = max(y for _, y in values)
        if clean2d and len(self) > 1 and max_y - min_y < .01 and max_x - min_x < 0.01:
            self.clean2d()
            min_x = min(x for x, _ in values)
            max_x = max(x for x, _ in values)
            min_y = min(y for _, y in values)
            max_y = max(y for _, y in values)

        bonds = self.__render_bonds()
        atoms, define, masks = self.__render_atoms(uid)
        if _embedding:
            return atoms, bonds, define, masks, uid, min_x, min_y, max_x, max_y

        font_size = _render_config['font_size']
        font125 = 1.25 * font_size
        _width = max_x - min_x + 4.0 * font_size
        _height = max_y - min_y + 2.5 * font_size
        viewbox_x = min_x - font125
        viewbox_y = -max_y - font125

        if width is None:
            width = f'{_width:.2f}cm'
        if height is None:
            height = f'{_height:.2f}cm'

        svg = [f'<svg width="{width}" height="{height}" '
               f'viewBox="{viewbox_x:.2f} {viewbox_y:.2f} {_width:.2f} {_height:.2f}" '
               'xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1">']
        svg.extend(_graph_svg(atoms, bonds, define, masks, uid, viewbox_x, viewbox_y, _width, _height))
        svg.append('</svg>')
        return '\n'.join(svg)

    @cached_method
    def _repr_svg_(self):
        return self.depict()

    def __render_bonds(self: Union['MoleculeContainer', 'DepictMolecule']):
        svg = []
        plane = self._plane
        double_space = _render_config['double_space']
        triple_space = _render_config['triple_space']
        wedge_space = _render_config['wedge_space']
        dash1, dash2 = _render_config['dashes']
        color = f' fill="{_render_config["bond_color"]}"'

        wedge = defaultdict(set)
        for n, m, s in self._wedge_map:
            wedge[n].add(m)
            wedge[m].add(n)

            nx, ny = plane[n]
            mx, my = plane[m]
            ny, my = -ny, -my
            dx, dy = _rotate_vector(0, wedge_space, mx - nx, ny - my)

            svg.append(f'      <path d="M{nx:.2f} {ny:.2f} L{mx + dx:.2f} {my + dy:.2f} '
                       f'L{mx - dx:.2f} {my - dy:.2f} Z"{s == 1 and color or ""}/>')

        for n, m, bond in self.bonds():
            if m in wedge[n]:
                continue
            order = bond.order
            nx, ny = plane[n]
            mx, my = plane[m]
            ny, my = -ny, -my
            if order in (1, 4):
                svg.append(f'      <line x1="{nx:.2f}" y1="{ny:.2f}" x2="{mx:.2f}" y2="{my:.2f}"/>')
            elif order == 2:
                dx, dy = _rotate_vector(0, double_space, mx - nx, ny - my)
                svg.append(f'      <line x1="{nx + dx:.2f}" y1="{ny - dy:.2f}" x2="{mx + dx:.2f}" y2="{my - dy:.2f}"/>')
                svg.append(f'      <line x1="{nx - dx:.2f}" y1="{ny + dy:.2f}" x2="{mx - dx:.2f}" y2="{my + dy:.2f}"/>')
            elif order == 3:
                dx, dy = _rotate_vector(0, triple_space, mx - nx, ny - my)
                svg.append(f'      <line x1="{nx + dx:.2f}" y1="{ny - dy:.2f}" x2="{mx + dx:.2f}" y2="{my - dy:.2f}"/>')
                svg.append(f'      <line x1="{nx:.2f}" y1="{ny:.2f}" x2="{mx:.2f}" y2="{my:.2f}"/>')
                svg.append(f'      <line x1="{nx - dx:.2f}" y1="{ny + dy:.2f}" x2="{mx - dx:.2f}" y2="{my + dy:.2f}"/>')
            else:
                svg.append(f'      <line x1="{nx:.2f}" y1="{ny:.2f}" x2="{mx:.2f}" y2="{my:.2f}" '
                           f'stroke-dasharray="{dash1:.2f} {dash2:.2f}"/>')

        for ring in self.aromatic_rings:
            cx = sum(plane[n][0] for n in ring) / len(ring)
            cy = sum(plane[n][1] for n in ring) / len(ring)

            for n, m in zip(ring, ring[1:]):
                nx, ny = plane[n]
                mx, my = plane[m]
                aromatic = _render_aromatic_bond(nx, ny, mx, my, cx, cy)
                if aromatic:
                    svg.append(aromatic)

            nx, ny = plane[ring[-1]]
            mx, my = plane[ring[0]]
            aromatic = _render_aromatic_bond(nx, ny, mx, my, cx, cy)
            if aromatic:
                svg.append(aromatic)
        return svg

    def __render_atoms(self: 'MoleculeContainer', uid):
        bonds = self._bonds
        plane = self._plane
        charges = self._charges
        radicals = self._radicals
        hydrogens = self._hydrogens

        carbon = _render_config['carbon']
        mapping = _render_config['mapping']
        span_size = _render_config['span_size']
        font_size = _render_config['font_size']
        monochrome = _render_config['monochrome']
        other_size = _render_config['other_size']
        atoms_colors = _render_config['atoms_colors']
        mapping_size = _render_config['mapping_size']
        dx_m, dy_m = _render_config['dx_m'], _render_config['dy_m']
        dx_ci, dy_ci = _render_config['dx_ci'], _render_config['dy_ci']
        symbols_font_style = _render_config['symbols_font_style']
        span_dy = _render_config['span_dy']
        other_font_style = _render_config['other_font_style']
        mapping_font_style = _render_config['mapping_font_style']

        if monochrome:
            map_fill = other_fill = 'black'
        else:
            map_fill = _render_config['mapping_color']
            other_fill = _render_config['other_color']

        font2 = .2 * font_size
        font3 = .3 * font_size
        font4 = .4 * font_size
        font5 = .5 * font_size
        font6 = .6 * font_size
        font7 = .7 * font_size
        font15 = .15 * font_size
        font25 = .25 * font_size
        stroke_width_s = font_size * .1
        stroke_width_o = other_size * .1
        stroke_width_m = mapping_size * .1

        # for cumulenes
        cumulenes = {y for x in self._cumulenes(heteroatoms=True) if len(x) > 2 for y in x[1:-1]}

        svg = []
        maps = []
        symbols = []
        fill_zone = []
        others = []
        define = []
        mask = []

        for n, atom in self._atoms.items():
            x, y = plane[n]
            y = -y
            symbol = atom.atomic_symbol
            if not bonds[n] or symbol != 'C' or carbon or charges[n] or radicals[n] or atom.isotope or n in cumulenes:
                if charges[n]:
                    others.append(f'        <text x="{x:.2f}" y="{y:.2f}" dx="{dx_ci:.2f}" dy="-{dy_ci:.2f}">'
                                  f'{_render_charge[charges[n]]}{"↑" if radicals[n] else ""}</text>')
                elif radicals[n]:
                    others.append(f'        <text x="{x:.2f}" y="{y:.2f}" dx="{dx_ci:.2f}" dy="-{dy_ci:.2f}">↑</text>')
                if atom.isotope:
                    others.append(f'        <text x="{x:.2f}" y="{y:.2f}" dx="-{dx_ci:.2f}" dy="-{dy_ci:.2f}" '
                                  f'text-anchor="end">{atom.isotope}</text>')

                if len(symbol) > 1:
                    dx = font7
                    dx_mm = dx_m + font5
                    if symbol[-1] in ('l', 'i', 'r', 't'):
                        rx = font6
                        ax = font25
                    else:
                        rx = font7
                        ax = font15
                    fill_zone.append(f'          <ellipse cx="{x - ax:.2f}" cy="{y:.2f}" rx="{rx}" ry="{font4}"/>')
                else:
                    if symbol == 'I':
                        dx = font15
                        dx_mm = dx_m
                    else:
                        dx = font4
                        dx_mm = dx_m + font2
                    fill_zone.append(f'          <circle cx="{x:.2f}" cy="{y:.2f}" r="{font4:.2f}"/>')

                h = hydrogens[n]
                if h == 1:
                    h = 'H'
                elif h:
                    h = f'H<tspan  dy="{span_dy:.2f}" font-size="{span_size:.2f}">{h}</tspan>'
                else:
                    h = ''
                symbols.append(f'        <text id="{uid}-{n}" x="{x:.2f}" y="{y:.2f}" dx="-{dx:.2f}" dy="{font4:.2f}">'
                               f'{symbol}{h}</text>')

                svg.append(f'      <use xlink:href="#{uid}-{n}" '
                           f'fill="{"black" if monochrome else atoms_colors[atom.atomic_number - 1]}"/>')

                if mapping:
                    maps.append(f'        <text x="{x:.2f}" y="{y:.2f}" dx="-{dx_mm:.2f}" '
                                f'dy="{dy_m + font3:.2f}">{n}</text>')
            elif mapping:
                maps.append(f'        <text x="{x:.2f}" y="{y:.2f}" dx="-{dx_m:.2f}" dy="{dy_m:.2f}">{n}</text>')

        if svg:  # group atoms symbols
            if fill_zone:
                mask.append('        <g>')
                mask.extend(fill_zone)
                mask.append('        </g>')

            svg.insert(0, f'    <g font-size="{font_size:.2f}" font-family="{symbols_font_style}">')
            svg.append('    </g>')
            define.append(f'      <g id="{uid}-symbols" font-size="{font_size:.2f}" '
                          f'font-family="{symbols_font_style}">')
            define.extend(symbols)
            define.append('      </g>')
            mask.append('        <g stroke="black">\n          '
                        f'<use xlink:href="#{uid}-symbols" stroke-width="{stroke_width_s:.2f}"/>')

            if others:
                define.append(f'      <g id="{uid}-attrs" font-size="{other_size:.2f}" '
                              f'font-family="{other_font_style}">')
                define.extend(others)
                define.append('      </g>')
                svg.append(f'    <use xlink:href="#{uid}-attrs" fill="{other_fill}"/>')
                mask.append(f'          <use xlink:href="#{uid}-attrs" stroke-width="{stroke_width_o:.2f}"/>')
        if maps:
            if not svg:  # no atoms but maps
                mask.append('        <g stroke="black">')
            define.append(f'      <g id="{uid}-mapping" font-size="{mapping_size:.2f}" '
                          f'font-family="{mapping_font_style}" text-anchor="end">')
            define.extend(maps)
            define.append('      </g>')
            svg.append(f'    <use xlink:href="#{uid}-mapping" fill="{map_fill}"/>')
            mask.append(f'          <use xlink:href="#{uid}-mapping" stroke-width="{stroke_width_m:.2f}"/>\n'
                        '        </g>')
        elif svg:  # no maps but atoms
            mask.append('        </g>')
        return svg, define, mask


class DepictReaction:
    __slots__ = ()

    def depict(self: 'ReactionContainer', *, width=None, height=None, clean2d: bool = True) -> str:
        """
        Depict reaction in SVG format.

        :param width: set svg width param. by default auto-calculated.
        :param height: set svg height param. by default auto-calculated.
        :param clean2d: calculate coordinates if necessary.
        """
        if not self._arrow:
            if clean2d:
                for m in self.molecules():
                    if len(m) > 1:
                        values = m._plane.values()  # noqa
                        min_x = min(x for x, _ in values)
                        max_x = max(x for x, _ in values)
                        min_y = min(y for _, y in values)
                        max_y = max(y for _, y in values)
                        if max_y - min_y < .01 and max_x - min_x < 0.01:
                            m.clean2d()
            self.fix_positions()

        r_atoms = []
        r_bonds = []
        r_defines = []
        r_masks = []
        r_uids = []
        r_max_x = r_max_y = r_min_y = 0
        for m in self.molecules():
            atoms, bonds, define, masks, uid, min_x, min_y, max_x, max_y = m.depict(clean2d=False, _embedding=True)
            r_atoms.append(atoms)
            r_bonds.append(bonds)
            r_defines.append(define)
            r_masks.append(masks)
            r_uids.append(uid)
            if max_x > r_max_x:
                r_max_x = max_x
            if max_y > r_max_y:
                r_max_y = max_y
            if min_y < r_min_y:
                r_min_y = min_y

        font_size = _render_config['font_size']
        font125 = 1.25 * font_size
        _width = r_max_x + 4.0 * font_size
        _height = r_max_y - r_min_y + 2.5 * font_size
        viewbox_x = -font125
        viewbox_y = -r_max_y - font125

        if width is None:
            width = f'{_width:.2f}cm'
        if height is None:
            height = f'{_height:.2f}cm'

        svg = [f'<svg width="{width}" height="{height}" '
               f'viewBox="{viewbox_x:.2f} {viewbox_y:.2f} {_width:.2f} {_height:.2f}" '
               'xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1">\n'
               '  <defs>\n    <marker id="arrow" markerWidth="10" markerHeight="10" '
               'refX="0" refY="3" orient="auto">\n      <path d="M0,0 L0,6 L9,3"/>\n    </marker>\n  </defs>\n'
               f'  <line x1="{self._arrow[0]:.2f}" y1="0" x2="{self._arrow[1]:.2f}" y2="0" '
               'fill="none" stroke="black" stroke-width=".04" marker-end="url(#arrow)"/>']

        sings_plus = self._signs
        if sings_plus:
            svg.append(f'  <g fill="none" stroke="black" stroke-width=".04">')
            for x in sings_plus:
                svg.append(f'    <line x1="{x + .35:.2f}" y1="0" x2="{x + .65:.2f}" y2="0"/>')
                svg.append(f'    <line x1="{x + .5:.2f}" y1="0.15" x2="{x + .5:.2f}" y2="-0.15"/>')
            svg.append('  </g>')

        for atoms, bonds, define, masks, uid in zip(r_atoms, r_bonds, r_defines, r_masks, r_uids):
            svg.extend(_graph_svg(atoms, bonds, define, masks, uid, viewbox_x, viewbox_y, _width, _height))
        svg.append('</svg>')
        return '\n'.join(svg)

    @cached_method
    def _repr_svg_(self):
        return self.depict()


__all__ = ['DepictMolecule', 'DepictReaction', 'depict_settings']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _rules():
    from ...containers import QueryContainer
    rules = []

    # Oxo-acid salts. Me-[O,S,Se]-[P,S,Cl,Se,Br,I,Si]=O
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), neighbors=2, hybridization=1)
    q.add_atom(ListElement(['C', 'N', 'P', 'S', 'Cl', 'Se', 'Br', 'I', 'Si']))
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # Thiophosphate salts.
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), neighbors=2, hybridization=1)
    q.add_atom('P')
    q.add_atom(ListElement(['S', 'Se']))
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # Phenole salts. Me-[O,S,Se]-Ar
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), neighbors=2, hybridization=1)
    q.add_atom(ListElement(['C', 'N']), hybridization=4)
    q.add_bond(1, 2, 1)
    rules.append(q)

    # Nitrate. Me-O-[N+](=O)[O-]
    q = QueryContainer()
    q.add_atom('O', neighbors=2, hybridization=1)
    q.add_atom('N', charge=1)
    q.add_atom('O', charge=-1)
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 2)
    rules.append(q)

    # halogenides and hydroxy
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'F', 'Cl', 'Br', 'I']), neighbors=1, hybridization=1)
    rules.append(q)
    return rules


def _acids():
    from ... import smiles

    tmp = ['Cl', 'Br', 'I', 'O[N+](=O)[O-]', 'ON=O', 'OP(O)(O)=O', 'COP(O)(=O)OC',
           'OS(O)(=O)=O', 'CS(O)(=O)=O', 'OS(=O)(=O)C(F)(F)F', 'CC1=CC=C(C=C1)S(O)(=O)=O',
           'OC(O)=O', 'CC(O)=O', 'OC(=O)C(F)(F)F', 'OCC(O)=O', 'CC(O)C(O)=O', 'OC(=O)C(O)=O', 'OC(=O)C(Cl)Cl',
           'OC(=O)C=CC(O)=O', 'OC(C(O)C(O)=O)C(O)=O',
           'O[Cl](=O)(=O)=O']
    acs = set()
    for x in tmp:
        x = smiles(x)
        x.thiele()
        acs.add(x)
    return acs


acids = Proxy(_acids)
rules = Proxy(_rules)


__all__ = ['acids', 'rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import count
from re import compile, findall, search
from .parser import parser
from .tokenize import smarts_tokenize
from ...containers import QueryContainer
from ...periodictable import QueryElement


cx_radicals = compile(r'\^[1-7]:[0-9]+(?:,[0-9]+)*')
cx_hh = compile(r'atomProp(:[0-9]+\.(?:hyb|het|msk)\.[0-9]+)+')
hybridization = {'4': 4, '3': 1, '2': 2, '1': 3}

# AD-HOC for masked atoms in SMARTS
# not thread safe
global_free_masked = count(10 ** 9 + 1)


def smarts(data: str):
    """
    Parse SMARTS string.

    * stereo ignored.
    * only D, a, h, r and !R atom primitives supported.
    * bond order list and not bond supported.
    * [not]ring bond supported only in combination with explicit bonds, not bonds and bonds orders lists.
    * mapping, charge and isotopes supported.
    * list of elements supported.
    * A - treats as any element. <A> primitive (aliphatic) ignored.
    * M - treats as any metal..
    * <&> logic operator unsupported.
    * <;> logic operator is mandatory except (however preferable) for charge, isotope, stereo marks.
    * CXSMARTS radicals supported.
    * hybridization and heteroatoms count in CXSMARTS atomProp notation coded as <hyb> and <het> keys.
    * masked atom - `chython.Reactor` specific mark for masking reactant atoms from deletion.
        Coded in CXSMARTS atomProp as <msk> key with any value.

    For example::

        [C;r5,r6;a]-;!@[C;h1,h2] |^1:1,atomProp:1.hyb.24:1.het.0| - aromatic C member of 5 or 6 atoms ring
        connected with non-ring single bond to aromatic or SP2 radical C with 1 or 2 hydrogens.

    Alternative hybridization, heteroatoms and masks coding:

    * primitive <xN> - heteroatoms (e.g. x2 - two heteroatoms)
    * primitive <zN> - hybridization (N = 1 - sp3, 2 - sp2, 3 - sp, 4 - aromatic)
    * primitive <M> - masked atom

    Note: atom numbers greater than 10 ** 9 forbidden for usage and reserved for masked atoms numbering.
    In multiprocess mode has potential bugs in reaction enumeration task then used templates prepared from components
    from different processes. For avoiding, prepare templates on single process and then share it.
    """
    smr, *cx = data.split()

    hyb = {}
    het = {}
    msk = []
    if cx and cx[0].startswith('|') and cx[0].endswith('|'):
        radicals = [int(x) for x in findall(cx_radicals, cx[0]) for x in x[3:].split(',')]

        if hh := search(cx_hh, cx[0]):
            for x in hh.group().split(':')[1:]:
                i, h, v = x.split('.')
                i = int(i)
                if h == 'hyb':
                    hyb[i] = [hybridization[x] for x in v]
                elif h == 'het':
                    het[i] = [int(y) for y in v]
                else:
                    msk.append(i)
    else:
        radicals = []

    data = parser(smarts_tokenize(smr), False)[0]

    for x in radicals:
        data['atoms'][x]['is_radical'] = True
    for i, v in hyb.items():
        data['atoms'][i]['hybridization'] = v
    for i, v in het.items():
        data['atoms'][i]['heteroatoms'] = v
    for i in msk:
        data['atoms'][i]['masked'] = True

    g = QueryContainer()

    mapping = {}
    free = count(max(a['mapping'] for a in data['atoms']) + 1)
    for i, a in enumerate(data['atoms']):
        mapping[i] = n = a.pop('mapping') or next(global_free_masked if a['masked'] else free)
        e = a.pop('element')
        if it := a.pop('isotope'):
            if isinstance(e, int):
                e = QueryElement.from_atomic_number(e)(it)
            else:
                e = QueryElement.from_symbol(e)(it)
        g.add_atom(e, n, **a)

    for n, m, b in data['bonds']:
        g.add_bond(mapping[n], mapping[m], b)
    return g


__all__ = ['smarts']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Alexander Nikanshin <17071996sasha@gmail.com>
#  Copyright 2019 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from .element import Element
from .groups import GroupXV
from .periods import PeriodII, PeriodIII, PeriodIV, PeriodV, PeriodVI, PeriodVII


class N(Element, PeriodII, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 7

    @property
    def isotopes_distribution(self):
        return {13: 0., 14: 0.99632, 15: 0.00368}

    @property
    def isotopes_masses(self):
        return {13: 13.005738, 14: 14.003074, 15: 15.000109}

    @property
    def _common_valences(self):
        return 3,

    @property
    def _valences_exceptions(self):
        return ((-1, False, 2, ()), (1, False, 4, ()),
                (0, True, 2, ()))  # *NO, etc

    @property
    def atomic_radius(self):
        return .56


class P(Element, PeriodIII, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 15

    @property
    def isotopes_distribution(self):
        return {31: 1.0, 32: 0., 33: 0.}

    @property
    def isotopes_masses(self):
        return {31: 30.973762, 32: 31.973908, 33: 32.971726}

    @property
    def _common_valences(self):
        return 3, 5

    @property
    def _valences_exceptions(self):
        return ((-1, False, 2, ()), (1, False, 4, ()),
                (-1, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))))

    @property
    def atomic_radius(self):
        return .98


class As(Element, PeriodIV, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 33

    @property
    def isotopes_distribution(self):
        return {75: 1.0, 76: 0., 77: 0.}

    @property
    def isotopes_masses(self):
        return {75: 74.921596, 76: 75.922394, 77: 76.920647}

    @property
    def _common_valences(self):
        return 0, 3, 5

    @property
    def _valences_exceptions(self):
        return (1, False, 4, ()),

    @property
    def atomic_radius(self):
        return 1.14


class Sb(Element, PeriodV, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 51

    @property
    def isotopes_distribution(self):
        return {121: 0.5721, 123: 0.4279}

    @property
    def isotopes_masses(self):
        return {121: 120.903818, 123: 122.904216}

    @property
    def _common_valences(self):
        return 0, 3, 5

    @property
    def _valences_exceptions(self):
        return ((1, False, 4, ()),
                (-1, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))))

    @property
    def atomic_radius(self):
        return 1.33


class Bi(Element, PeriodVI, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 83

    @property
    def isotopes_distribution(self):
        return {207: 0., 209: 1.0, 210: 0.}

    @property
    def isotopes_masses(self):
        return {207: 206.978471, 209: 208.980383, 210: 209.984120}

    @property
    def _common_valences(self):
        return 0, 3

    @property
    def _valences_exceptions(self):
        return ((3, False, 0, ()),
                (0, False, 0, ((1, 'Cl'),)),
                (0, False, 0, ((1, 'Br'),)),

                (0, False, 0, ((1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((1, 'Br'), (1, 'Br'))),
                (0, False, 0, ((1, 'I'), (1, 'I'))),
                (0, False, 0, ((1, 'S'), (1, 'S'))),
                (0, False, 0, ((2, 'S'),)),
                (0, False, 0, ((1, 'Se'), (1, 'Se'))),
                (0, False, 0, ((2, 'Se'),)),

                (0, False, 0, ((1, 'Cl'), (1, 'Cl'), (1, 'Cl'), (1, 'Cl'))),
                (0, False, 0, ((2, 'O'), (2, 'O'))),

                (0, False, 0, ((1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'), (1, 'F'))),
                (0, False, 0, ((1, 'O'), (2, 'O'), (2, 'O'))),
                (0, False, 0, ((1, 'O'), (1, 'O'), (1, 'O'), (2, 'O'))))

    @property
    def atomic_radius(self):
        return 1.43


class Mc(Element, PeriodVII, GroupXV):
    __slots__ = ()

    @property
    def atomic_number(self):
        return 115

    @property
    def isotopes_distribution(self):
        return {289: 1.0}

    @property
    def isotopes_masses(self):
        return {289: 289.0}

    @property
    def _common_valences(self):
        return 0,

    @property
    def _valences_exceptions(self):
        return ()

    @property
    def atomic_radius(self):
        return 1.43  # unknown, taken radius of previous element in group


__all__ = ['N', 'P', 'As', 'Sb', 'Bi', 'Mc']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Dict, Generic, Iterator, Optional, Tuple, TypeVar
from ..algorithms.isomorphism import Isomorphism
from ..algorithms.morgan import Morgan
from ..algorithms.rings import Rings
from ..exceptions import AtomNotFound, MappingError, BondNotFound


Atom = TypeVar('Atom')
Bond = TypeVar('Bond')


class Graph(Generic[Atom, Bond], Morgan, Rings, Isomorphism, ABC):
    __slots__ = ('_atoms', '_bonds', '_charges', '_radicals', '__dict__', '__weakref__')
    __class_cache__ = {}

    _atoms: Dict[int, Atom]
    _bonds: Dict[int, Dict[int, Bond]]
    _charges: Dict[int, int]
    _radicals: Dict[int, bool]

    def __init__(self):
        self._atoms = {}
        self._bonds = {}
        self._charges = {}
        self._radicals = {}

    def atom(self, n: int) -> Atom:
        return self._atoms[n]

    def has_atom(self, n: int) -> bool:
        return n in self._atoms

    def atoms(self) -> Iterator[Tuple[int, Atom]]:
        """
        iterate over all atoms
        """
        return iter(self._atoms.items())

    @property
    def atoms_count(self) -> int:
        return len(self._atoms)

    @property
    def atoms_numbers(self) -> Iterator[int]:
        return iter(self._atoms)

    def bond(self, n: int, m: int) -> Bond:
        try:
            return self._bonds[n][m]
        except KeyError as e:
            raise BondNotFound from e

    def has_bond(self, n: int, m: int) -> bool:
        try:
            self._bonds[n]  # check if atom exists
            return n in self._bonds[m]
        except KeyError:
            raise AtomNotFound

    def bonds(self) -> Iterator[Tuple[int, int, Bond]]:
        """
        iterate other all bonds
        """
        seen = set()
        for n, m_bond in self._bonds.items():
            seen.add(n)
            for m, bond in m_bond.items():
                if m not in seen:
                    yield n, m, bond

    @cached_property
    def bonds_count(self) -> int:
        return sum(len(x) for x in self._bonds.values()) // 2

    @abstractmethod
    def add_atom(self, atom: Atom, n: Optional[int] = None, *, charge: int = 0, is_radical: bool = False) -> int:
        """
        new atom addition
        """
        if n is None:
            n = max(self._atoms, default=0) + 1
        elif not isinstance(n, int):
            raise TypeError('mapping should be integer')
        elif n in self._atoms:
            raise MappingError('atom with same number exists')
        elif not isinstance(is_radical, bool):
            raise TypeError('bool expected')
        elif not isinstance(charge, int):
            raise TypeError('formal charge should be int in range [-4, 4]')
        elif charge > 4 or charge < -4:
            raise ValueError('formal charge should be in range [-4, 4]')

        atom._attach_graph(self, n)
        self._atoms[n] = atom
        self._charges[n] = charge
        self._radicals[n] = is_radical
        self._bonds[n] = {}
        self.__dict__.clear()
        return n

    @abstractmethod
    def add_bond(self, n: int, m: int, bond: Bond):
        """
        Add bond.
        """
        if n == m:
            raise MappingError('atom loops impossible')
        if n not in self._bonds or m not in self._bonds:
            raise AtomNotFound('atoms not found')
        if n in self._bonds[m]:
            raise MappingError('atoms already bonded')

        self._bonds[n][m] = self._bonds[m][n] = bond
        self.__dict__.clear()

    @abstractmethod
    def copy(self):
        """
        copy of graph
        """
        copy = object.__new__(self.__class__)
        copy._charges = self._charges.copy()
        copy._radicals = self._radicals.copy()

        copy._atoms = ca = {}
        for n, atom in self._atoms.items():
            atom = atom.copy()
            ca[n] = atom
            atom._attach_graph(copy, n)
        return copy

    @abstractmethod
    def union(self, other: 'Graph'):
        """
        Merge Graphs into one.
        """
        u = self.copy()
        u._charges.update(other._charges)
        u._radicals.update(other._radicals)

        ua = u._atoms
        for n, atom in other._atoms.items():
            ua[n] = atom = atom.copy()
            atom._attach_graph(u, n)
        return u

    def flush_cache(self):
        self.__dict__.clear()

    def __copy__(self):
        return self.copy()

    def __or__(self, other):
        """
        G | H is union of graphs
        """
        return self.union(other)

    def __len__(self):
        return len(self._atoms)

    def __iter__(self) -> Iterator[int]:
        return iter(self._atoms)

    def __bool__(self):
        return bool(self._atoms)

    def __getstate__(self):
        state = {'atoms': self._atoms, 'bonds': self._bonds, 'charges': self._charges,
                 'radicals': self._radicals}
        from chython import pickle_cache

        if pickle_cache:
            state['cache'] = {k: v for k, v in self.__dict__.items() if k != '__cached_method___hash__'}
        return state

    def __setstate__(self, state):
        self._atoms = state['atoms']
        for n, a in state['atoms'].items():
            a._attach_graph(self, n)
        self._charges = state['charges']
        self._radicals = state['radicals']
        self._bonds = state['bonds']
        if 'cache' in state:
            self.__dict__.update(state['cache'])


__all__ = ['Graph']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque, defaultdict
from itertools import product
from typing import TYPE_CHECKING
from ..aromatics.kekule import _kekule_component
from ...exceptions import InvalidAromaticRing


if TYPE_CHECKING:
    from chython import MoleculeContainer


class HeteroArenes:
    __slots__ = ()

    def _enumerate_hetero_arene_tautomers(self: 'MoleculeContainer'):
        atoms = self._atoms
        bonds = self._bonds
        hydrogens = self._hydrogens
        charges = self._charges
        radicals = self._radicals

        rings = defaultdict(list)  # aromatic skeleton
        for n, m_bond in bonds.items():
            for m, bond in m_bond.items():
                if bond.order == 4:
                    rings[n].append(m)
        if not rings:
            return

        acceptors = set()
        donors = set()
        single_bonded = set()
        for n, ms in rings.items():
            if len(ms) == 2:
                if atoms[n].atomic_number in (5, 7, 15):
                    if not charges[n] and not radicals[n]:
                        # only neutral B, N, P
                        if hydrogens[n]:  # pyrrole
                            donors.add(n)
                        elif len(bonds[n]) == 2:  # pyridine
                            acceptors.add(n)
                        else:
                            single_bonded.add(n)
                elif charges[n] == -1 and atoms[n].atomic_number == 6:  # ferrocene
                    single_bonded.add(n)
            elif len(ms) == 3 and atoms[n].atomic_number in (5, 7, 15) and not charges[n] and not radicals[n]:
                single_bonded.add(n)
        if not donors or not acceptors:
            return

        atoms = set(rings)
        components = []
        while atoms:
            start = atoms.pop()
            component = {start: rings[start]}
            queue = deque([start])
            while queue:
                current = queue.popleft()
                for n in rings[current]:
                    if n not in component:
                        queue.append(n)
                        component[n] = rings[n]

            atoms.difference_update(component)
            if donors.isdisjoint(component) or acceptors.isdisjoint(component):
                continue
            components.append(component)

        if not components:
            return
        for component in components:
            for d, a in product(component.keys() & donors, component.keys() & acceptors):
                sb = component.keys() & single_bonded
                sb.add(a)  # now pyrrole
                try:
                    next(_kekule_component(component, sb, (), 0))
                except InvalidAromaticRing:
                    continue
                mol = self.copy()
                mol._hydrogens[d] = 0
                mol._hydrogens[a] = 1
                yield mol


__all__ = ['HeteroArenes']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2021 Dmitrij Zanadvornykh <zandmitrij@gmail.com>
#  Copyright 2018 Tagir Akhmetshin <tagirshin@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from typing import List, TYPE_CHECKING, Union, Tuple
from ._charged import fixed_rules, morgan_rules
from ._groups import *
from ._metal_organics import rules as metal_rules
from ...containers.bonds import Bond
from ...exceptions import ValenceError, ImplementationError
from ...periodictable import H


if TYPE_CHECKING:
    from chython import MoleculeContainer


class Standardize:
    __slots__ = ()

    def canonicalize(self: 'MoleculeContainer', *, fix_tautomers=True, keep_kekule=False,
                     logging=False, ignore=True) -> Union[bool, List[Tuple[Tuple[int, ...], int, str]]]:
        """
        Convert molecule to canonical forms of functional groups and aromatic rings without explicit hydrogens.

        :param logging: return log.
        :param ignore: ignore standardization bugs.
        :param fix_tautomers: convert tautomers to canonical forms.
        :param keep_kekule: return kekule form.
        """
        k = self.kekule()
        s = self.standardize(_fix_stereo=False, logging=True, ignore=ignore, fix_tautomers=fix_tautomers)
        h, changed = self.implicify_hydrogens(_fix_stereo=False, logging=True)

        if keep_kekule:  # save bond orders
            bonds = [(b, b.order) for _, _, b in self.bonds()]

        t = self.thiele(fix_tautomers=fix_tautomers)
        c = self.standardize_charges(prepare_molecule=False, logging=True)

        if keep_kekule and t:  # restore
            if c:  # we need to do full kekule again
                self.kekule()
            else:
                for b, o in bonds:
                    b._Bond__order = o  # noqa
            self.flush_cache()

        if logging:
            if k:
                s.insert(0, ((), -1, 'kekulized'))
            if h:
                s.append((tuple(changed), -1, 'implicified'))
            if t:
                s.append(((), -1, 'aromatized'))
            if c:
                s.append((tuple(c), -1, 'recharged'))
            return s
        return bool(k or s or h or t or c)

    def standardize(self: Union['MoleculeContainer', 'Standardize'], *, logging=False, ignore=True, fix_tautomers=True,
                    _fix_stereo=True) -> Union[bool, List[Tuple[Tuple[int, ...], int, str]]]:
        """
        Standardize functional groups. Return True if any non-canonical group found.

        :param fix_tautomers: convert tautomers to canonical forms.
        :param logging: return list of fixed atoms with matched rules.
        :param ignore: ignore standardization bugs.
        """
        r = self.fix_resonance(logging=True, _fix_stereo=False)
        if r:
            log = [(tuple(r), -1, 'resonance fixed')]
            fixed = set(r)
        else:
            log, fixed = [], set()

        l, f = self.__standardize(double_rules, fix_tautomers)
        log.extend(l)
        fixed.update(f)
        if f:
            l, f = self.__standardize(double_rules, fix_tautomers)  # double shot rules for overlapped groups
            log.extend(l)
            fixed.update(f)
        l, f = self.__standardize(single_rules, fix_tautomers)
        log.extend(l)
        fixed.update(f)
        l, f = self.__standardize(metal_rules, fix_tautomers)  # metal-organics fix
        log.extend(l)
        fixed.update(f)

        if b := fixed.intersection(n for n, h in self._hydrogens.items() if h is None):
            if ignore:
                log.append((tuple(b), -1, 'standardization failed'))
            else:
                raise ImplementationError(f'standardization leads to invalid valences: {b}')

        if fixed:
            self.flush_cache()
            if _fix_stereo:
                self.fix_stereo()

        if logging:
            if fixed:
                log.append((tuple(fixed), -1, 'standardized atoms'))
            return log
        return bool(fixed)

    def standardize_charges(self: 'MoleculeContainer', *, logging=False, prepare_molecule=True,
                            _fix_stereo=True) -> Union[bool, List[int]]:
        """
        Set canonical positions of charges in heterocycles and some nitrogen compounds.

        :param logging: return list of changed atoms.
        :param prepare_molecule: do thiele procedure.
        """
        changed: List[int] = []
        bonds = self._bonds
        hydrogens = self._hydrogens
        charges = self._charges

        if prepare_molecule:
            self.thiele()

        seen = set()
        # not morgan
        for q, fix in fixed_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                match = set(mapping.values())
                if len(match.intersection(seen)) > 2:  # if matched more than 2 atoms
                    continue
                seen.update(match)
                # if not 2 neighbors and 1 hydrogen  or 3 neighbors within 1st and second atoms - break
                atom_1, atom_2 = mapping[1], mapping[2]
                if len(bonds[atom_1]) == 2:
                    if not hydrogens[atom_1]:
                        continue
                elif all(x == 4 for x in bonds[atom_1].values()):
                    continue

                if len(bonds[atom_2]) == 2:
                    if not hydrogens[atom_2]:
                        continue
                elif all(x == 4 for x in bonds[atom_2].values()):
                    continue

                if fix:
                    atom_3 = mapping[3]
                    charges[atom_3] = 0
                    changed.append(atom_3)
                else:
                    charges[atom_1] = 0
                    changed.append(atom_1)
                charges[atom_2] = 1
                changed.append(atom_2)  # add atoms to changed

        # morgan
        pairs = []
        for q, fix in morgan_rules:
            for mapping in q.get_mapping(self, automorphism_filter=False):
                match = set(mapping.values())
                if len(match.intersection(seen)) > 2:  # if matched more than 2 atoms
                    continue
                seen.update(match)
                atom_1, atom_2 = mapping[1], mapping[2]
                if len(bonds[atom_1]) == 2:
                    if not hydrogens[atom_1]:
                        continue
                elif all(x == 4 for x in bonds[atom_1].values()):
                    continue

                if len(bonds[atom_2]) == 2:
                    if not hydrogens[atom_2]:
                        continue
                elif all(x == 4 for x in bonds[atom_2].values()):
                    continue

                if fix:
                    atom_3 = mapping[3]
                    charges[atom_3] = 0
                    changed.append(atom_3)
                else:
                    # remove charge from 1st N atom
                    charges[atom_1] = 0
                pairs.append((atom_1, atom_2, fix))

        if pairs:
            self.__dict__.pop('atoms_order', None)  # remove cached morgan
            for atom_1, atom_2, fix in pairs:
                if self.atoms_order[atom_1] > self.atoms_order[atom_2]:
                    charges[atom_2] = 1
                    changed.append(atom_2)
                    if not fix:
                        changed.append(atom_1)
                else:
                    charges[atom_1] = 1
                    if fix:
                        changed.append(atom_1)
            del self.__dict__['atoms_order']  # remove invalid morgan
        if changed:
            self.flush_cache()  # clear cache
            if _fix_stereo:
                self.fix_stereo()
            if logging:
                return changed
            return True
        if logging:
            return []
        return False

    def remove_hydrogen_bonds(self: 'MoleculeContainer', *, keep_to_terminal=True, _fix_stereo=True) -> int:
        """Remove hydrogen bonds marked with 8 (any) bond

        :param keep_to_terminal: Keep any bonds to terminal hydrogens
        :return: removed bonds count
        """
        bonds = self._bonds
        hg = defaultdict(set)
        for n, atom in self._atoms.items():
            if atom.atomic_number == 1:
                for m, b in bonds[n].items():
                    if b.order == 8:
                        hg[n].add(m)

        if keep_to_terminal:
            for n, ms in hg.items():
                if len(bonds[n]) == len(ms):  # H~A or A~H~A etc case
                    m = ms.pop()
                    if m in hg:  # H~H case
                        hg[m].discard(n)

        seen = set()
        c = 0
        for n, ms in hg.items():
            seen.add(n)
            for m in ms:
                if m in seen:
                    continue
                c += 1
                del bonds[n][m], bonds[m][n]
        if c:
            self.flush_cache()
            if _fix_stereo:
                self.fix_stereo()
        return c

    def implicify_hydrogens(self: 'MoleculeContainer', *, logging=False, _fix_stereo=True) -> \
            Union[int, Tuple[int, List[int]]]:
        """
        Remove explicit hydrogen if possible. Return number of removed hydrogens.
        Works only with Kekule forms of aromatic structures.
        Keeps isotopes of hydrogen.

        :param logging: return list of changed atoms.
        """
        atoms = self._atoms
        charges = self._charges
        radicals = self._radicals
        bonds = self._bonds
        plane = self._plane
        hydrogens = self._hydrogens
        parsed_mapping = self._parsed_mapping
        calc_implicit = self._calc_implicit

        explicit = defaultdict(list)
        for n, atom in atoms.items():
            if atom.atomic_number == 1 and (atom.isotope is None or atom.isotope == 1):
                if len(bonds[n]) > 1:
                    raise ValenceError(f'Hydrogen atom {{{n}}} has invalid valence. Try to use remove_hydrogen_bonds()')
                for m, b in bonds[n].items():
                    if b.order == 1:
                        if atoms[m].atomic_number != 1:  # not H-H
                            explicit[m].append(n)
                    elif b.order != 8:
                        raise ValenceError(f'Hydrogen atom {{{n}}} has invalid valence {{{b.order}}}.')

        to_remove = set()
        fixed = []
        for n, hs in explicit.items():
            atom = atoms[n]
            charge = charges[n]
            is_radical = radicals[n]
            len_h = len(hs)
            for i in range(len_h, 0, -1):
                hi = hs[:i]
                explicit_sum = 0
                explicit_dict = defaultdict(int)
                for m, bond in bonds[n].items():
                    if m not in hi and bond.order != 8:
                        explicit_sum += bond.order
                        explicit_dict[(bond.order, atoms[m].atomic_number)] += 1
                try:
                    rules = atom.valence_rules(charge, is_radical, explicit_sum)
                except ValenceError:
                    break
                if any(s.issubset(explicit_dict) and all(explicit_dict[k] >= c for k, c in d.items()) and h >= i
                       for s, d, h in rules):
                    to_remove.update(hi)
                    fixed.append(n)
                    break

        for n in to_remove:
            del atoms[n]
            del charges[n]
            del radicals[n]
            del plane[n]
            del hydrogens[n]
            for m in bonds.pop(n):
                del bonds[m][n]
            try:
                del parsed_mapping[n]
            except KeyError:
                pass

        for n in fixed:
            calc_implicit(n)

        if to_remove:
            self.flush_cache()
            self._conformers = [{x: y for x, y in c.items() if x not in to_remove} for c in self._conformers]  # noqa
            if _fix_stereo:
                self.fix_stereo()

        if logging:
            return len(to_remove), fixed
        return len(to_remove)

    def explicify_hydrogens(self: 'MoleculeContainer', *, start_map=None, _return_map=False, _fix_stereo=True) -> \
            Union[int, List[Tuple[int, int]]]:
        """
        Add explicit hydrogens to atoms.

        :return: number of added atoms
        """
        hydrogens = self._hydrogens
        to_add = []
        for n, h in hydrogens.items():
            try:
                to_add.extend([n] * h)
            except TypeError:
                raise ValenceError(f'atom {{{n}}} has valence error')

        if to_add:
            log = []
            bonds = self._bonds
            m = start_map
            for n in to_add:
                m = self.add_atom(H(), m)
                bonds[n][m] = bonds[m][n] = b = Bond(1)
                b._attach_graph(self, n, m)
                hydrogens[n] = 0
                log.append((n, m))
                m += 1

            if _fix_stereo:
                self.fix_stereo()
            if _return_map:
                return log
            return len(to_add)
        elif _return_map:
            return []
        return 0

    def check_valence(self: 'MoleculeContainer') -> List[int]:
        """
        Check valences of all atoms.

        :return: list of invalid atoms
        """
        return [n for n, h in self._hydrogens.items() if h is None]  # only invalid atoms have None hydrogens.

    def clean_isotopes(self: 'MoleculeContainer') -> bool:
        """
        Clean isotope marks from molecule.
        Return True if any isotope found.
        """
        atoms = self._atoms
        isotopes = [x for x in atoms.values() if x.isotope]
        if isotopes:
            for i in isotopes:
                i._Core__isotope = None
            self.flush_cache()
            self.fix_stereo()
            return True
        return False

    def __standardize(self: 'MoleculeContainer', rules, fix_tautomers):
        bonds = self._bonds
        charges = self._charges
        radicals = self._radicals
        calc_implicit = self._calc_implicit

        log = []
        fixed = set()
        flush = False
        for r, (pattern, atom_fix, bonds_fix, any_atoms, is_tautomer) in enumerate(rules):
            if not fix_tautomers and is_tautomer:
                continue
            hs = set()
            seen = set()
            for mapping in pattern.get_mapping(self, automorphism_filter=False):
                match = set(mapping.values())
                if not match.isdisjoint(seen):  # skip intersected groups
                    continue
                if any_atoms:  # accept overlapping of Any-atoms
                    seen.update(match - {mapping[n] for n in any_atoms})
                else:
                    seen.update(match)
                for n, (ch, ir) in atom_fix.items():
                    n = mapping[n]
                    hs.add(n)
                    charges[n] += ch
                    if charges[n] > 4:
                        charges[n] -= ch
                        log.append((tuple(match), r, f'bad charge formed. changes omitted: {pattern}'))
                        break  # skip changes
                    if ir is not None:
                        radicals[n] = ir
                else:
                    for n, m, b in bonds_fix:
                        n = mapping[n]
                        m = mapping[m]
                        hs.add(n)
                        hs.add(m)
                        if m in bonds[n]:
                            bonds[n][m]._Bond__order = b  # noqa
                            if b == 8:
                                # expected original molecule don't contain `any` bonds or these bonds not changed
                                flush = True
                        else:
                            if b != 8:
                                flush = True
                            bonds[n][m] = bonds[m][n] = b = Bond(b)
                            b._attach_graph(self, n, m)
                    log.append((tuple(match), r, str(pattern)))

            if not hs:  # not matched
                continue
            # flush cache only for changed atoms.
            if flush:  # neighbors count changed
                ngb = self.__dict__['__cached_args_method_neighbors']
                for n in hs:
                    try:
                        del ngb[(n,)]
                    except KeyError:
                        pass
                del self.__dict__['bonds_count']
                flush = False
            # need hybridization recalculation
            hyb = self.__dict__['__cached_args_method_hybridization']
            for n in hs:
                try:
                    del hyb[(n,)]
                except KeyError:  # already flushed before
                    pass
            for n in hs:  # hydrogens count recalculation
                calc_implicit(n)
            del self.__dict__['_cython_compiled_structure']
            fixed.update(hs)
        return log, fixed


__all__ = ['Standardize']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy


def _rules():
    from ... import smarts

    rules = []

    # Aromatic N-Oxide
    #
    #  : N :  >>  : [N+] :
    #    \\           \
    #     O           [O-]
    #
    q = smarts('[N;a;D3:1]=[O;D1:2]')
    atom_fix = {1: 1, 2: -1}
    bonds_fix = ((1, 2, 1),)
    # query, atom fix, bond fix, allow multimatch
    rules.append((q, atom_fix, bonds_fix, False))

    # Aromatic N-Nitride?
    #
    #  : N :  >>  : [N+] :
    #    \\           \
    #     N           [N-]
    #
    q = smarts('[N;a;D3:1]=[N;D1,D2;z2:2]')
    atom_fix = {1: 1, 2: -1}
    bonds_fix = ((1, 2, 1),)
    rules.append((q, atom_fix, bonds_fix, False))

    #
    # : [S+] : >> : S :
    #    |          \\
    #   [O-]         O
    #
    q = smarts('[S;a;D3;+:1]-[O;D1;-:2]')
    atom_fix = {1: 0, 2: 0}
    bonds_fix = ((1, 2, 2),)
    rules.append((q, atom_fix, bonds_fix, False))

    #
    # [O-]-N:C:C:[N+]=O
    #
    q = smarts('[N;a;D3;+:1](=[O;D1:2]):[C:3]:[C:4]:[N;D3:5]-[O;D1;-:6]')
    atom_fix = {5: 1, 2: -1}
    bonds_fix = ((1, 2, 1),)
    rules.append((q, atom_fix, bonds_fix, False))

    #
    # N : A : N - ?
    #  :     :
    #   C # C
    q = smarts('[C;a;D2:1]:1#[C;D2:2]:[N;D2,D3:3]:[C,N:4]:[N;D2:5]:1')
    atom_fix = {}
    bonds_fix = ((1, 2, 4),)
    rules.append((q, atom_fix, bonds_fix, False))

    #
    # C:[N+]:[C-]
    #    \\
    #     O
    #
    q = smarts('[N;a;D3;+:1](=[O;D1:2])(:[C;D2,D3;-:3]):[C;D2,D3:4]')
    atom_fix = {2: -1, 3: 0}
    bonds_fix = ((1, 2, 1),)
    rules.append((q, atom_fix, bonds_fix, False))

    #
    #  O=[N+] : C
    #     :     :
    #    O : N : C
    q = smarts('[N;a;D3;+:1]:1(=[O;D1:2]):[O;D2:3]:[N;D2,D3:4]:[C;D2,D3:5]:[C;D2,D3:6]:1')
    atom_fix = {}
    bonds_fix = ((1, 3, 1), (1, 6, 1), (3, 4, 1), (4, 5, 1), (5, 6, 2))
    rules.append((q, atom_fix, bonds_fix, False))

    # bad complex representation
    #  :           :
    #   N - [M]  >> N ... [M]
    #  :           :
    q = smarts('[N;a;D3:1]-[M:2]')
    atom_fix = {}
    bonds_fix = ((1, 2, 8),)
    rules.append((q, atom_fix, bonds_fix, True))
    return rules


rules = Proxy(_rules)


__all__ = ['rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import chain, repeat
from typing import List, Tuple, TYPE_CHECKING, Union
from ._groups import *


if TYPE_CHECKING:
    from chython import ReactionContainer


class GroupsFix:
    __slots__ = ()

    def fix_groups_mapping(self: 'ReactionContainer', *, logging: bool = False) -> \
            Union[bool, List[Tuple[str, Tuple[int, ...]]]]:
        """
        Fix atom-to-atom mapping of some functional groups. Return True if found AAM errors.
        """
        if not self:
            if logging:
                return []
            return False

        log = []
        seen = set()
        remap = {}
        pamer = {}
        r_groups = set()
        p_groups = set()
        r_subs = set()
        p_subs = set()

        # find xonyl groups. any charged-neutral combinations
        for pattern in xonyl_groups:
            for m, g in chain(zip(self.reactants, repeat(r_groups)), zip(self.products, repeat(p_groups))):
                atoms = m._atoms
                for mapping in pattern.get_mapping(m, automorphism_filter=False):
                    n1, n2, n3 = mapping[1], mapping[2], mapping[3]
                    if (t := atoms[n2].atomic_number) == atoms[n3].atomic_number:
                        g.add((n1, n2, n3, atoms[n1].atomic_number, t))

        for pattern, _map in substituents_groups:
            for m, g in chain(zip(self.reactants, repeat(r_subs)), zip(self.products, repeat(p_subs))):
                atoms = m._atoms
                for mapping in pattern.get_mapping(m, automorphism_filter=False):
                    g.add((n := mapping[1], atoms[n].atomic_number,
                           tuple((n := mapping[x], y - 2, atoms[n].atomic_number) for x, y in _map), m))

        r_groups = list(r_groups)
        p_groups = list(p_groups)

        # find pairs
        if r_groups and p_groups:
            for n1, n2, n3, x1, x2 in r_groups:
                if n1 in seen:  # already remapped
                    continue
                for i, (m1, m2, m3, y1, y2) in enumerate(p_groups):
                    if m1 not in seen and n1 == m1 and x1 == y1 and x2 == y2:  # found pair
                        if n2 == m3 and n3 == m2:  # found switch
                            remap[m2] = m3
                            remap[m3] = m2
                            seen.add(n1)
                        break
                else:
                    continue
                del p_groups[i]

        if not p_groups:  # optimize
            r_subs.clear()

        # hydrolysis, etc.
        for (n1, x1, _map, m), g, r in chain(zip(r_subs, repeat(p_groups), repeat(remap)),
                                             zip(p_subs, repeat(r_groups), repeat(pamer))):
            if n1 in seen:
                continue
            for i, (m1, *m23, y1, y2) in enumerate(g):
                if m1 not in seen and n1 == m1 and x1 == y1:  # found center
                    if len(_map) == 1:  # acids substitutions.
                        ni, _, xi = _map[0]
                        # second neighbor should be disconnected from central atom.
                        if xi == y2 and m23[0] == ni and (m23[1] not in m._atoms or m23[1] not in m._bonds[n1]):
                            m2, m3 = m23
                            r[m2] = m3
                            r[m3] = m2
                            seen.add(n1)
                            break
                    elif all(xi == y2 and m23[mi] == ni for ni, mi, xi in _map):
                        m2, m3 = m23
                        r[m2] = m3
                        r[m3] = m2
                        seen.add(n1)
                        break
            else:
                continue
            del g[i]

        if remap:
            seen = set(remap)
            for m in self.products:
                if not seen.isdisjoint(m):
                    m.remap(remap)
            log.append(('products groups remapped', tuple(remap)))
        if pamer:
            seen = set(pamer)
            for m in self.reactants:
                if not seen.isdisjoint(m):
                    m.remap(pamer)
            log.append(('reactants groups remapped', tuple(pamer)))

        if log:
            self.flush_cache()
            if logging:
                return log
            return True
        elif logging:
            return []
        return False


__all__ = ['GroupsFix']



# -*- coding: utf-8 -*-
#
#  Copyright 2014-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from itertools import product
from ..containers import MoleculeContainer, QueryContainer
from ..containers.bonds import Bond
from ..periodictable import Element, ListElement, AnyElement


class BaseReactor:
    def __init__(self, reactants, products, delete_atoms, fix_rings, fix_tautomers):
        self.__to_delete = reactants.difference(products) if delete_atoms else ()

        # prepare atoms patch
        self.__elements = elements = {}
        self.__hydrogens = hydrogens = {}
        self.__variable = variable = []

        atoms = defaultdict(dict)
        for n, atom in products.atoms():
            atoms[n].update(charge=atom.charge, is_radical=atom.is_radical)
            if atom.atomic_number:  # replace atom
                elements[n] = Element.from_atomic_number(atom.atomic_number)(atom.isotope)
                if n not in reactants and isinstance(products, MoleculeContainer):
                    atoms[n]['xy'] = atom.xy
                    if atom.hybridization == 4 and atom.implicit_hydrogens is not None:
                        hydrogens[n] = atom.implicit_hydrogens  # keep it for new aromatic atoms
            elif n not in reactants:
                if not isinstance(atom, ListElement):
                    raise ValueError('New atom should be defined')
                elements[n] = [Element.from_symbol(x)() for x in atom._elements]
                variable.append(n)
            else:  # use atom from reactant
                if not isinstance(atom, AnyElement):
                    raise ValueError('Only AnyElement can be used for matched atom propagation')
                elements[n] = None

        if isinstance(products, QueryContainer):
            bonds = []
            for n, m, b in products.bonds():
                if len(b.order) > 1:
                    raise ValueError('bond list in patch not supported')
                else:
                    bonds.append((n, m, Bond(b.order[0])))
        else:
            bonds = [(n, m, b.copy()) for n, m, b in products.bonds()]

        self.__bonds = bonds
        self.__atom_attrs = dict(atoms)
        self.__products = products
        self.__fix_rings = fix_rings
        self.__fix_tautomers = fix_tautomers

    def _patcher(self, structure: MoleculeContainer, mapping):
        elements = self.__elements
        variable = self.__variable

        new = self.__prepare_skeleton(structure, mapping)
        self.__set_stereo(new, structure, mapping)

        if not variable:
            if self.__fix_rings:
                new.kekule()  # keeps stereo as is
                if not new.thiele(fix_tautomers=self.__fix_tautomers):  # fixes stereo if any ring aromatized
                    new.fix_stereo()
            else:
                new.fix_stereo()
            yield new
        else:
            copy = new.copy()
            if self.__fix_rings:
                copy.kekule()
                if not copy.thiele(fix_tautomers=self.__fix_tautomers):
                    copy.fix_stereo()
            else:
                copy.fix_stereo()
            yield copy

            for atoms in product(*(elements[x][1:] for x in variable)):
                copy = new.copy()
                for n, atom in zip(variable, atoms):
                    n = mapping[n]
                    # replace atom
                    copy._atoms[n] = a = atom.copy()  # noqa
                    a._attach_graph(copy, n)  # noqa
                    copy._calc_implicit(n)  # noqa
                if self.__fix_rings:
                    copy.kekule()
                    if not copy.thiele(fix_tautomers=self.__fix_tautomers):
                        copy.fix_stereo()
                    else:
                        copy.fix_stereo()
                else:
                    copy.fix_stereo()
                yield copy

    def __prepare_skeleton(self, structure, mapping):
        elements = self.__elements
        patch_hydrogens = self.__hydrogens
        patch_bonds = self.__bonds
        variable = self.__variable

        atoms = structure._atoms
        plane = structure._plane
        bonds = structure._bonds
        charges = structure._charges
        radicals = structure._radicals
        hydrogens = structure._hydrogens

        to_delete = {mapping[x] for x in self.__to_delete}
        if to_delete:
            # if deleted atoms have another path to remain fragment, the path is preserved
            remain = set(mapping.values()).difference(to_delete)
            delete, global_seen = set(), set()
            for x in to_delete:
                for n in bonds[x]:
                    if n in global_seen or n in remain:
                        continue
                    seen = {n}
                    global_seen.add(n)
                    stack = [x for x in bonds[n] if x not in global_seen]
                    while stack:
                        current = stack.pop()
                        if current in remain:
                            break
                        if current in to_delete:
                            continue
                        seen.add(current)
                        global_seen.add(current)
                        stack.extend([x for x in bonds[current] if x not in global_seen])
                    else:
                        delete.update(seen)

            to_delete.update(delete)

        new = structure.__class__()
        keep_hydrogens = {}
        max_atom = max(atoms)
        for n, atom in self.__atom_attrs.items():
            if n in mapping:  # add matched atoms
                m = mapping[n]
                e = elements[n]
                if e is None:
                    e = atoms[m]
                new.add_atom(e.copy(), m, xy=plane[m], _skip_hydrogen_calculation=True, **atom)
            else:  # new atoms
                max_atom += 1
                if n in variable:
                    # use first from the list
                    mapping[n] = new.add_atom(elements[n][0].copy(), max_atom, _skip_hydrogen_calculation=True, **atom)
                else:
                    mapping[n] = new.add_atom(elements[n].copy(), max_atom, _skip_hydrogen_calculation=True, **atom)
                    if n in patch_hydrogens:  # keep patch aromatic atoms hydrogens count
                        keep_hydrogens[max_atom] = patch_hydrogens[n]

        patch_atoms = set(new)  # don't move!
        for n, atom in structure.atoms():  # add unmatched atoms
            if n not in patch_atoms and n not in to_delete:
                new.add_atom(atom.copy(), n, charge=charges[n], is_radical=radicals[n], xy=plane[n],
                             _skip_hydrogen_calculation=True)
                keep_hydrogens[n] = hydrogens[n]  # keep hydrogens on unmatched atoms as is.

        for n, m, bond in patch_bonds:  # add patch bonds
            new.add_bond(mapping[n], mapping[m], bond.copy(), _skip_hydrogen_calculation=True)

        for n, m_bond in bonds.items():
            if n in to_delete:  # atoms for removing
                continue
            to_delete.add(n)  # reuse to_delete set for seen atoms
            for m, bond in m_bond.items():
                # ignore deleted atoms and patch atoms
                if m in to_delete or n in patch_atoms and m in patch_atoms:
                    continue
                new.add_bond(n, m, bond.copy(), _skip_hydrogen_calculation=True)

        # fix hydrogens count.
        new._hydrogens.update(keep_hydrogens)  # noqa
        for n in new:
            if n not in keep_hydrogens:
                new._calc_implicit(n)  # noqa
        return new

    def __set_stereo(self, new, structure, mapping):
        products = self.__products
        stereo_override = set()
        r_mapping = {m: n for n, m in mapping.items()}

        # set patch atoms stereo
        for n, s in products._atoms_stereo.items():
            m = mapping[n]
            new._atoms_stereo[m] = products._translate_tetrahedron_sign(n, [r_mapping[x] for x in
                                                                            new._stereo_tetrahedrons[m]], s)
            stereo_override.add(m)

        for n, s in products._allenes_stereo.items():
            m = mapping[n]
            t1, t2, *_ = new._stereo_allenes[m]
            new._allenes_stereo[m] = products._translate_allene_sign(n, r_mapping[t1], r_mapping[t2], s)
            stereo_override.add(m)

        for (n, m), s in products._cis_trans_stereo.items():
            nm = (mapping[n], mapping[m])
            try:
                t1, t2, *_ = new._stereo_cis_trans[nm]
            except KeyError:
                nm = nm[::-1]
                t2, t1, *_ = new._stereo_cis_trans[nm]
            new._cis_trans_stereo[nm] = products._translate_cis_trans_sign(n, m, r_mapping[t1], r_mapping[t2], s)
            stereo_override.update(nm)

        # set unmatched part stereo and not overridden by patch.
        for n, s in structure._atoms_stereo.items():
            if n in stereo_override or n not in new._stereo_tetrahedrons or \
                    new._bonds[n].keys() != structure._bonds[n].keys():
                # skip atoms with changed neighbors
                continue
            new._atoms_stereo[n] = structure._translate_tetrahedron_sign(n, new._stereo_tetrahedrons[n], s)

        for n, s in structure._allenes_stereo.items():
            if n in stereo_override or n not in new._stereo_allenes or \
                    set(new._stereo_allenes[n]) != set(structure._stereo_allenes[n]):
                # skip changed allenes
                continue
            t1, t2, *_ = new._stereo_allenes[n]
            new._allenes_stereo[n] = structure._translate_allene_sign(n, t1, t2, s)

        for nm, s in structure._cis_trans_stereo.items():
            n, m = nm
            if n in stereo_override or m in stereo_override:
                continue
            env = structure._stereo_cis_trans[nm]
            try:
                new_env = new._stereo_cis_trans[nm]
            except KeyError:
                nm = nm[::-1]
                try:
                    new_env = new._stereo_cis_trans[nm]
                except KeyError:
                    continue
                t2, t1, *_ = new_env
            else:
                t1, t2, *_ = new_env
            if set(env) != set(new_env):
                continue
            new._cis_trans_stereo[nm] = structure._translate_cis_trans_sign(n, m, t1, t2, s)


__all__ = ['BaseReactor']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport sqrt


cdef extern from "Python.h":
    dict _PyDict_NewPresized(Py_ssize_t minused)


@cython.boundscheck(False)
@cython.wraparound(False)
def possible_bonds(double[:, ::1] xyz not None, double[::1] radii not None, double multiplier):
    cdef unsigned int size, max_bonds, c = 0, b = 0, n, m, k
    cdef double d, nx, ny, nz, rn, mx, my, mz
    cdef dict py_bonds, tmp

    size = xyz.shape[0]
    max_bonds = size * 10  # each atom has less then 10 neighbors approximately

    cdef unsigned int *ns = <unsigned int *> PyMem_Malloc(size * sizeof(unsigned int))
    cdef unsigned int *mi = <unsigned int *> PyMem_Malloc(max_bonds * sizeof(unsigned int))
    cdef double *ds = <double *> PyMem_Malloc(max_bonds * sizeof(double))

    if not ns or not mi or not ds:
        raise MemoryError()

    for n in range(size - 1):
        nx, ny, nz = xyz[n, 0], xyz[n, 1], xyz[n, 2]
        rn = radii[n]
        for m in range(n + 1, size):
            mx, my, mz = nx - xyz[m, 0], ny - xyz[m, 1], nz - xyz[m, 2]
            d = sqrt(mx * mx + my * my + mz * mz)
            if d <= (rn + radii[m]) * multiplier:
                mi[c] = m
                ds[c] = d
                c += 1
        ns[n] = c

    # prepare dict of dicts
    py_bonds = _PyDict_NewPresized(size)
    for n in range(size):
        py_bonds[n + 1] = {}

    for n in range(size - 1):
        c = ns[n]
        n += 1
        tmp = py_bonds[n]
        for m in range(b, c):
            k = mi[m] + 1
            tmp[k] = py_bonds[k][n] = ds[m]
        b = c

    PyMem_Free(ns)
    PyMem_Free(mi)
    PyMem_Free(ds)
    return py_bonds



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import defaultdict
from functools import cached_property
from typing import Dict, Optional, Tuple, TYPE_CHECKING, Union


if TYPE_CHECKING:
    from chython import MoleculeContainer, QueryContainer
    Container = Union[MoleculeContainer, QueryContainer]


_heteroatoms = {5, 6, 7, 8, 14, 15, 16, 17, 33, 34, 35, 52, 53}

# 1  2
#  \ |
#   \|
#    n---3
#   /
#  /
# 0
_tetrahedron_translate = {(0, 1, 2): False, (1, 2, 0): False, (2, 0, 1): False,
                          (0, 2, 1): True, (1, 0, 2): True, (2, 1, 0): True,
                          (0, 3, 1): False, (3, 1, 0): False, (1, 0, 3): False,
                          (0, 1, 3): True, (1, 3, 0): True, (3, 0, 1): True,
                          (0, 2, 3): False, (2, 3, 0): False, (3, 0, 2): False,
                          (0, 3, 2): True, (3, 2, 0): True, (2, 0, 3): True,
                          (1, 3, 2): False, (3, 2, 1): False, (2, 1, 3): False,
                          (1, 2, 3): True, (2, 3, 1): True, (3, 1, 2): True}
# 2       1
#  \     /
#   n---m
#  /     \
# 0       3
_alkene_translate = {(0, 1): False, (1, 0): False, (0, 3): True, (3, 0): True,
                     (2, 3): False, (3, 2): False, (2, 1): True, (1, 2): True}

# allowed atoms. these atoms have stable covalent bonds.
_organic_subset = {1, 5, 6, 7, 8, 9, 14, 15, 16, 17, 33, 34, 35, 52, 53}


class Stereo:
    __slots__ = ()

    @cached_property
    def cumulenes(self) -> Tuple[Tuple[int, ...], ...]:
        """
        Alkenes, allenes and cumulenes atoms numbers.
        """
        return tuple(self._cumulenes())

    @cached_property
    def tetrahedrons(self: 'Container') -> Tuple[int, ...]:
        """
        Carbon sp3 atoms numbers.
        """
        atoms = self._atoms
        bonds = self._bonds
        charges = self._charges
        radicals = self._radicals

        tetra = []
        for n, atom in atoms.items():
            if atom.atomic_number == 6 and not charges[n] and not radicals[n]:
                env = bonds[n]
                if all(int(x) == 1 for x in env.values()):
                    if sum(int(x) for x in env.values()) > 4:
                        continue
                    tetra.append(n)
        return tuple(tetra)

    def clean_stereo(self: 'Container'):
        """
        Remove stereo data.
        """
        self._atoms_stereo.clear()
        self._allenes_stereo.clear()
        self._cis_trans_stereo.clear()
        self.flush_cache()

    def get_mapping(self: 'Container', other: 'Container', **kwargs):
        atoms_stereo = self._atoms_stereo
        allenes_stereo = self._allenes_stereo
        cis_trans_stereo = self._cis_trans_stereo
        if atoms_stereo or allenes_stereo or cis_trans_stereo:
            other_atoms_stereo = other._atoms_stereo
            other_allenes_stereo = other._allenes_stereo
            other_cis_trans_stereo = other._cis_trans_stereo
            other_translate_tetrahedron_sign = other._translate_tetrahedron_sign
            other_translate_allene_sign = other._translate_allene_sign
            other_translate_cis_trans_sign = other._translate_cis_trans_sign

            tetrahedrons = self._stereo_tetrahedrons
            cis_trans = self._stereo_cis_trans
            allenes = self._stereo_allenes

            for mapping in super().get_mapping(other, **kwargs):
                for n, s in atoms_stereo.items():
                    m = mapping[n]
                    if m not in other_atoms_stereo:  # self stereo atom not stereo in other
                        break
                    # translate stereo mark in other in order of self tetrahedron
                    if other_translate_tetrahedron_sign(m, [mapping[x] for x in tetrahedrons[n]]) != s:
                        break
                else:
                    for n, s in allenes_stereo.items():
                        m = mapping[n]
                        if m not in other_allenes_stereo:  # self stereo allene not stereo in other
                            break
                        # translate stereo mark in other in order of self allene
                        nn, nm, *_ = allenes[n]
                        if other_translate_allene_sign(m, mapping[nn], mapping[nm]) != s:
                            break
                    else:
                        for nm, s in cis_trans_stereo.items():
                            n, m = nm
                            on, om = mapping[n], mapping[m]
                            if (on, om) not in other_cis_trans_stereo:
                                if (om, on) not in other_cis_trans_stereo:
                                    break  # self stereo cis_trans not stereo in other
                                else:
                                    nn, nm, *_ = cis_trans[nm]
                                    if other_translate_cis_trans_sign(om, on, mapping[nm], mapping[nn]) != s:
                                        break
                            else:
                                nn, nm, *_ = cis_trans[nm]
                                if other_translate_cis_trans_sign(on, om, mapping[nn], mapping[nm]) != s:
                                    break
                        else:
                            yield mapping
        else:
            yield from super().get_mapping(other, **kwargs)

    def _translate_tetrahedron_sign(self: 'Container', n, env, s=None):
        """
        Get sign of chiral tetrahedron atom for specified neighbors order

        :param n: stereo atom
        :param env: neighbors order
        :param s: if None, use existing sign else translate given to molecule
        """
        if s is None:
            s = self._atoms_stereo[n]

        order = self._stereo_tetrahedrons[n]
        if len(order) == 3:
            if len(env) == 4:  # hydrogen atom passed to env
                atoms = self._atoms
                # hydrogen always last in order
                try:
                    order = (*order, next(x for x in env if atoms[x].atomic_number == 1))  # see translate scheme
                except StopIteration:
                    raise KeyError
            elif len(env) != 3:  # pyramid or tetrahedron expected
                raise ValueError('invalid atoms list')
        elif len(env) not in (3, 4):  # pyramid or tetrahedron expected
            raise ValueError('invalid atoms list')

        translate = tuple(order.index(x) for x in env[:3])
        if _tetrahedron_translate[translate]:
            return not s
        return s

    def _translate_cis_trans_sign(self: 'Container', n, m, nn, nm, s=None):
        """
        Get sign for specified opposite neighbors

        :param n: first double bonded atom
        :param m: last double bonded atom
        :param nn: neighbor of first atom
        :param nm: neighbor of last atom
        :param s: if None, use existing sign else translate given to molecule
        """
        if s is None:
            try:
                s = self._cis_trans_stereo[(n, m)]
            except KeyError:
                s = self._cis_trans_stereo[(m, n)]
                n, m = m, n  # in alkenes sign not order depended
                nn, nm = nm, nn

        atoms = self._atoms
        n0, n1, n2, n3 = self._stereo_cis_trans[(n, m)]
        if nn == n0:  # same start
            t0 = 0
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n1:
            t0 = 1
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        elif nn == n2 or n2 is None and atoms[nn].atomic_number == 1:
            t0 = 2
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n3 or n3 is None and atoms[nn].atomic_number == 1:
            t0 = 3
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        else:
            raise KeyError

        if _alkene_translate[(t0, t1)]:
            return not s
        return s

    def _translate_allene_sign(self: 'Container', c, nn, nm, s=None):
        """
        get sign for specified opposite neighbors

        :param c: central double bonded atom
        :param nn: neighbor of first double bonded atom
        :param nm: neighbor of last double bonded atom
        :param s: if None, use existing sign else translate given to molecule
        """
        if s is None:
            s = self._allenes_stereo[c]

        atoms = self._atoms
        n0, n1, n2, n3 = self._stereo_allenes[c]
        if nn == n0:  # same start
            t0 = 0
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n1:
            t0 = 1
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        elif nn == n2 or n2 is None and atoms[nn].atomic_number == 1:
            t0 = 2
            if nm == n1:
                t1 = 1
            elif nm == n3 or n3 is None and atoms[nm].atomic_number == 1:
                t1 = 3
            else:
                raise KeyError
        elif nn == n3 or n3 is None and atoms[nn].atomic_number == 1:
            t0 = 3
            if nm == n0:
                t1 = 0
            elif nm == n2 or n2 is None and atoms[nm].atomic_number == 1:
                t1 = 2
            else:
                raise KeyError
        else:
            raise KeyError

        if _alkene_translate[(t0, t1)]:
            return not s
        return s

    def _cumulenes(self: 'Container', heteroatoms=False):
        atoms = self._atoms
        bonds = self._bonds

        adj = defaultdict(set)  # double bonds adjacency matrix
        if heteroatoms:
            for n, atom in atoms.items():
                if atom.atomic_number in _heteroatoms:
                    adj_n = adj[n].add
                    for m, bond in bonds[n].items():
                        if int(bond) == 2 and atoms[m].atomic_number in _heteroatoms:
                            adj_n(m)
        else:
            for n, atom in atoms.items():
                if atom.atomic_number == 6:
                    adj_n = adj[n].add
                    for m, bond in bonds[n].items():
                        if int(bond) == 2 and atoms[m].atomic_number == 6:
                            adj_n(m)
        if not adj:
            return ()

        terminals = [x for x, y in adj.items() if len(y) == 1]
        cumulenes = []
        while terminals:
            n = terminals.pop(0)
            m = adj[n].pop()
            path = [n, m]
            while m not in terminals:
                adj_m = adj[m]
                if len(adj_m) > 2:  # not cumulene. SO3 etc.
                    cumulenes.extend(zip(path, path[1:]))  # keep single double bonds.
                    break
                adj_m.discard(n)
                n, m = m, adj_m.pop()
                path.append(m)
            else:
                terminals.remove(m)
                adj[m].pop()
                cumulenes.append(tuple(path))
        return cumulenes

    @cached_property
    def _stereo_cumulenes(self: 'Container') -> Dict[Tuple[int, ...], Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Cumulenes which contains at least one non-hydrogen neighbor on both ends
        """
        # 5       4
        #  \     /
        #   2---3
        #  /     \
        # 1       6
        bonds = self._bonds
        atoms = self._atoms
        cumulenes = {}
        for path in self.cumulenes:
            nf = bonds[path[0]]
            nl = bonds[path[-1]]
            n1, m1 = path[1], path[-2]
            if any(b not in (1, 4) and atoms[m].atomic_number not in _organic_subset for m, b in nf.items() if m != n1)\
                    or any(b not in (1, 4) and atoms[m].atomic_number not in _organic_subset
                           for m, b in nl.items() if m != m1):
                continue  # skip X=C=C structures and metal-carbon complexes
            nn = [x for x in nf if x != n1 and atoms[x].atomic_number != 1]
            mn = [x for x in nl if x != m1 and atoms[x].atomic_number != 1]
            if nn and mn:
                sn = nn[1] if len(nn) == 2 else None
                sm = mn[1] if len(mn) == 2 else None
                cumulenes[path] = (nn[0], mn[0], sn, sm)
        return cumulenes

    @cached_property
    def _stereo_tetrahedrons(self: 'Container') -> Dict[int, Union[Tuple[int, int, int], Tuple[int, int, int, int]]]:
        """
        Tetrahedrons which contains at least 3 non-hydrogen neighbors
        """
        #    2
        #    |
        # 1--K--3
        #    |
        #    4?
        atoms = self._atoms
        bonds = self._bonds
        tetrahedrons = {}
        for n in self.tetrahedrons:
            if any(atoms[x].atomic_number not in _organic_subset for x in bonds[n]):
                continue  # skip metal-carbon complexes
            env = tuple(x for x in bonds[n] if atoms[x].atomic_number != 1)
            if len(env) in (3, 4):
                tetrahedrons[n] = env
        return tetrahedrons

    @cached_property
    def _stereo_cis_trans(self) -> Dict[Tuple[int, int], Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Cis-trans bonds which contains at least one non-hydrogen neighbor on both ends
        """
        return {(n, m): env for (n, *mid, m), env in self._stereo_cumulenes.items() if not len(mid) % 2}

    @cached_property
    def _stereo_cis_trans_paths(self) -> Dict[Tuple[int, int], Tuple[int, ...]]:
        return {(path[0], path[-1]): path for path in self._stereo_cumulenes if not len(path) % 2}

    @cached_property
    def _stereo_cis_trans_terminals(self) -> Dict[int, Tuple[int, int]]:
        """
        Cis-Trans terminal atoms to cis-trans key mapping
        """
        terminals = {}
        for nm in self._stereo_cis_trans_paths:
            n, m = nm
            terminals[n] = terminals[m] = nm
        return terminals

    @cached_property
    def _stereo_allenes(self) -> Dict[int, Tuple[int, int, Optional[int], Optional[int]]]:
        """
        Allenes which contains at least one non-hydrogen neighbor on both ends
        """
        return {path[len(path) // 2]: env for path, env in self._stereo_cumulenes.items() if len(path) % 2}

    @cached_property
    def _stereo_allenes_centers(self) -> Dict[int, int]:
        """
        Allene terminal atom to center mapping
        """
        terminals = {}
        for c, (n, m) in self._stereo_allenes_terminals.items():
            terminals[n] = terminals[m] = c
        return terminals

    @cached_property
    def _stereo_allenes_terminals(self) -> Dict[int, Tuple[int, int]]:
        """
        Allene center atom to terminals mapping
        """
        return {c: (path[0], path[-1]) for c, path in self._stereo_allenes_paths.items()}

    @cached_property
    def _stereo_allenes_paths(self) -> Dict[int, Tuple[int, ...]]:
        return {path[len(path) // 2]: path for path in self._stereo_cumulenes if len(path) % 2}


__all__ = ['Stereo']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _stripped_rules():
    from ...containers import QueryContainer
    rules = []

    # Oxo-acid salts. [O,S,Se-]-[C,N,P,S,Cl,Se,Br,I,Si]=O
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), charge=-1)
    q.add_atom(ListElement(['C', 'N', 'P', 'S', 'Cl', 'Se', 'Br', 'I', 'Si']))
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # Thiophosphate salts.
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), charge=-1)
    q.add_atom('P')
    q.add_atom(ListElement(['S', 'Se']))
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # Phenole salts, alcoholates. [O,S,Se-]-Ar
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'Se']), charge=-1)
    q.add_atom(ListElement(['C', 'N']))
    q.add_bond(1, 2, 1)
    rules.append(q)

    # Nitrate. [O-]-[N+](=O)[O-]
    q = QueryContainer()
    q.add_atom('O', charge=-1)
    q.add_atom('N', charge=1)
    q.add_atom('O', charge=-1)
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 2)
    rules.append(q)

    # ions
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'F', 'Cl', 'Br', 'I', 'S']), charge=-1, neighbors=0)
    rules.append(q)
    return rules


def _rules():
    from ...containers import QueryContainer
    rules = _stripped_rules()

    # Guanidine
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C')
    q.add_atom('N')
    q.add_atom('N')
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 1)
    rules.append(q)

    # Oxo-guanidine, Amino-guanidine
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('C')
    q.add_atom('N')
    q.add_atom('N')
    q.add_atom(ListElement(['O', 'N']))
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 1)
    q.add_bond(1, 5, 1)
    rules.append(q)

    # O-alkyl-isourea, S-alkyl-isothiaurea
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C')
    q.add_atom('N')
    q.add_atom(ListElement(['O', 'S']), neighbors=2, hybridization=1, heteroatoms=0)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 1)
    rules.append(q)

    # Dialkyl imidocarbonate
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C')
    q.add_atom('O', neighbors=2, heteroatoms=0)
    q.add_atom('O', neighbors=2, heteroatoms=0)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(2, 4, 1)
    rules.append(q)

    # Amidine
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C', heteroatoms=2)
    q.add_atom('N')
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # O-alkyl-imidate (oxazoline)
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C', heteroatoms=2)
    q.add_atom('O', neighbors=2, heteroatoms=0)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # Amidoxime. O-N=C([C,H])N
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('C', heteroatoms=2)
    q.add_atom('N')
    q.add_atom('O')
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    q.add_bond(1, 4, 1)
    rules.append(q)

    # Oxime, Hydrazone. [O,N]-N=C([C,H])[C,H]
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('C', heteroatoms=1, hybridization=2)
    q.add_atom(ListElement(['N', 'O']))
    q.add_bond(1, 2, 2)
    q.add_bond(1, 3, 1)
    rules.append(q)

    # Imine. [C,H]N=C([C,H])[C,H]
    q = QueryContainer()
    q.add_atom('N', heteroatoms=0)
    q.add_atom('C', heteroatoms=1, hybridization=2)
    q.add_bond(1, 2, 2)
    rules.append(q)

    # Alkyl amine, Hydroxylamine, Hydrazine
    q = QueryContainer()
    q.add_atom('N', neighbors=1)
    q.add_atom(ListElement(['C', 'O', 'N']), hybridization=1, heteroatoms=1)
    q.add_bond(1, 2, 1)
    rules.append(q)

    # Dialkyl amine, Alkyl hydroxylamine, Alkyl hydrazine
    q = QueryContainer()
    q.add_atom('N', neighbors=2)
    q.add_atom(ListElement(['C', 'O', 'N']), hybridization=1, heteroatoms=1)
    q.add_atom('C', hybridization=1, heteroatoms=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 1)
    rules.append(q)

    # Trialkyl amine, Dialkyl-hydroxylamine, Dialkyl-hydrazine
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom(ListElement(['C', 'O', 'N']), hybridization=1, heteroatoms=1)
    q.add_atom('C', hybridization=1, heteroatoms=1)
    q.add_atom('C', hybridization=1, heteroatoms=1)
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 1)
    q.add_bond(1, 4, 1)
    rules.append(q)

    # Pyridine. Imidazole. Triazole. :N:
    q = QueryContainer()
    q.add_atom('N', hybridization=4, hydrogens=0, neighbors=2)
    rules.append(q)
    return rules


stripped_rules = Proxy(_stripped_rules)
rules = Proxy(_rules)


__all__ = ['stripped_rules', 'rules']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy
from ...periodictable import ListElement


def _sugar_group():
    from ...containers import QueryContainer
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'N']), hydrogens=(1, 2))  # enol
    q.add_atom(ListElement(['O', 'N']))  # ketone
    q.add_atom('C', hybridization=1)
    q.add_atom('C', hybridization=2)
    q.add_bond(1, 3, 1)
    q.add_bond(3, 4, 1)
    q.add_bond(2, 4, 2)
    return q


def _keto_rules():
    from ...containers import QueryContainer
    rules = []
    # first atom is H-acceptor
    # second is direction

    # C-C=[O,S,NH]
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'N']), neighbors=1)
    q.add_atom('C', heteroatoms=1, hybridization=2, neighbors=(2, 3))
    q.add_bond(1, 2, 2)
    rules.append(q)

    # C-C=N-[C,N,O]
    q = QueryContainer()
    q.add_atom('N', neighbors=2)
    q.add_atom('C', heteroatoms=1, hybridization=2, neighbors=(2, 3))
    q.add_atom(ListElement(['C', 'N', 'O']))
    q.add_bond(1, 2, 2)
    q.add_bond(1, 3, 1)
    rules.append(q)

    # [C,H]-N=N-C
    q = QueryContainer()
    q.add_atom('N', neighbors=(1, 2), heteroatoms=1)
    q.add_atom('N')
    q.add_atom('C', heteroatoms=1)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # [S,O,NR;H]-C=N
    q = QueryContainer()
    q.add_atom('N')
    q.add_atom('C')
    q.add_atom(ListElement(['N', 'S', 'O']), hydrogens=1)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # [NH2]-C=N-R
    q = QueryContainer()
    q.add_atom('N', neighbors=2)
    q.add_atom('C')
    q.add_atom('N', neighbors=1)
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # S=C-[N,O;H]
    q = QueryContainer()
    q.add_atom('S', neighbors=1)
    q.add_atom('C')
    q.add_atom(ListElement(['N', 'O']), hydrogens=(1, 2))
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)

    # O=C-[N,S;H]
    q = QueryContainer()
    q.add_atom('O')
    q.add_atom('C')
    q.add_atom(ListElement(['N', 'S']), hydrogens=(1, 2))
    q.add_bond(1, 2, 2)
    q.add_bond(2, 3, 1)
    rules.append(q)
    return rules


def _enol_rules():
    from ...containers import QueryContainer
    rules = []
    # first atom is H-donor
    # second is direction

    # C=C-[OH,SH,NH2]
    q = QueryContainer()
    q.add_atom(ListElement(['O', 'S', 'N']), neighbors=1)
    q.add_atom('C', heteroatoms=1, hybridization=2, neighbors=(2, 3))
    q.add_bond(1, 2, 1)
    rules.append(q)

    # C=C-[NH]-[C,N,O]
    q = QueryContainer()
    q.add_atom('N', neighbors=2)
    q.add_atom('C', heteroatoms=1, hybridization=2, neighbors=(2, 3))
    q.add_atom(ListElement(['C', 'N', 'O']))
    q.add_bond(1, 2, 1)
    q.add_bond(1, 3, 1)
    rules.append(q)

    # [C,H]-[NH]-N=C
    q = QueryContainer()
    q.add_atom('N', neighbors=(1, 2), hybridization=1, heteroatoms=1)
    q.add_atom('N')
    q.add_atom('C', heteroatoms=1)
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # [S,O,NR;H]-C=N
    q = QueryContainer()
    q.add_atom(ListElement(['N', 'S', 'O']), hydrogens=1)
    q.add_atom('C')
    q.add_atom('N')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # [NH2]-C=N-R
    q = QueryContainer()
    q.add_atom('N', neighbors=1)
    q.add_atom('C')
    q.add_atom('N', neighbors=2)
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # S=C-[N,O;H]
    q = QueryContainer()
    q.add_atom(ListElement(['N', 'O']), hydrogens=(1, 2))
    q.add_atom('C')
    q.add_atom('S', neighbors=1)
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)

    # O=C-[N,S;H]
    q = QueryContainer()
    q.add_atom(ListElement(['N', 'S']), hydrogens=(1, 2))
    q.add_atom('C')
    q.add_atom('O')
    q.add_bond(1, 2, 1)
    q.add_bond(2, 3, 2)
    rules.append(q)
    return rules


keto_rules = Proxy(_keto_rules)
enol_rules = Proxy(_enol_rules)
sugar_group = Proxy(_sugar_group)


__all__ = ['keto_rules', 'enol_rules', 'sugar_group']



# -*- coding: utf-8 -*-
#
#  Copyright 2018-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from abc import abstractmethod
from array import array
from collections import defaultdict, deque
from functools import cached_property, partial
from itertools import permutations
from typing import Any, Collection, Dict, Iterator, Optional, TYPE_CHECKING, Union
from .._functions import lazy_product
from ..periodictable.element import Element, Query


if TYPE_CHECKING:
    from chython.containers.graph import Graph


frequency = {1: 10,  # H
             6: 9,  # C
             8: 8,  # O
             7: 7,  # N
             15: 6, 16: 6,  # P, S
             9: 5,  # F
             17: 4, 35: 4,  # Cl, Br
             53: 3,  # I
             5: 2, 14: 2,  # B, Si
             11: 1, 12: 1, 19: 1, 20: 1}  # Na, Mg, K, Ca


def atom_frequency(x):
    return frequency.get(x.atomic_number, 0)


class Isomorphism:
    __slots__ = ()

    def __lt__(self, other):
        if len(self) >= len(other):
            return False
        return self.is_substructure(other)

    def __le__(self, other):
        return self.is_substructure(other)

    def __gt__(self, other):
        if len(self) <= len(other):
            return False
        return other.is_substructure(self)

    def __ge__(self, other):
        return other.is_substructure(self)

    def __contains__(self: 'Graph', other: Union[Element, Query, str]):
        """
        Atom in Structure test.
        """
        if isinstance(other, str):
            return any(other == x.atomic_symbol for x in self._atoms.values())
        return any(other == x for x in self._atoms.values())

    def is_substructure(self, other, /) -> bool:
        """
        Test self is substructure of other
        """
        try:
            next(self.get_mapping(other, automorphism_filter=False))
        except StopIteration:
            return False
        return True

    def is_equal(self, other, /) -> bool:
        """
        Test self is same structure as other
        """
        if len(self) != len(other):
            return False
        try:
            next(self.get_mapping(other, automorphism_filter=False))
        except StopIteration:
            return False
        return True

    @abstractmethod
    def get_mapping(self, other, /, *, automorphism_filter: bool = True,
                    searching_scope: Optional[Collection[int]] = None, _cython=False) -> Iterator[Dict[int, int]]:
        """
        Get self to other substructure mapping generator.

        :param automorphism_filter: Skip matches to same atoms.
        :param searching_scope: substructure atoms list to localize isomorphism.
        """
        if _cython:  # ad-hoc for QueryContainer
            try:  # windows? ;)
                from ._isomorphism import get_mapping as _cython_get_mapping
            except ImportError:
                components, closures = self._compiled_query
                get_mapping = partial(_get_mapping, query_closures=closures, o_atoms=other._atoms, o_bonds=other._bonds)
            else:
                components = self._cython_compiled_query  # override to cython data

                def get_mapping(query, scope):
                    return _cython_get_mapping(*query, *other._cython_compiled_structure,
                                               array('I', [n in scope for n in other]))
        else:
            components, closures = self._compiled_query
            get_mapping = partial(_get_mapping, query_closures=closures, o_atoms=other._atoms, o_bonds=other._bonds)

        if searching_scope is not None and not isinstance(searching_scope, set):
            searching_scope = set(searching_scope)

        seen = set()
        if len(components) == 1:
            for candidate in other._connected_components:
                if searching_scope:
                    candidate = searching_scope.intersection(candidate)
                    if not candidate:
                        continue
                for mapping in get_mapping(components[0], scope=candidate):
                    if automorphism_filter:
                        atoms = frozenset(mapping.values())
                        if atoms in seen:
                            continue
                        seen.add(atoms)
                    yield mapping
        else:
            for candidates in permutations(other._connected_components, len(components)):
                mappers = []
                for component, candidate in zip(components, candidates):
                    if searching_scope:
                        candidate = searching_scope.intersection(candidate)
                        if not candidate:
                            break
                    mappers.append(get_mapping(component, scope=candidate))
                else:
                    for match in lazy_product(*mappers):
                        mapping = match[0].copy()
                        for m in match[1:]:
                            mapping.update(m)
                        if automorphism_filter:
                            atoms = frozenset(mapping.values())
                            if atoms in seen:
                                continue
                            seen.add(atoms)
                        yield mapping

    @cached_property
    def _compiled_query(self: 'Graph'):
        components, closures = _compile_query(self._atoms, self._bonds,
                                              {n: atom_frequency(a) for n, a in self._atoms.items()})
        if self.connected_components_count > 1:
            order = {x: n for n, c in enumerate(self.connected_components) for x in c}
            components.sort(key=lambda x: order[x[0][0]])
        return components, closures

    def is_automorphic(self):
        """
        Test for automorphism symmetry of graph.
        """
        try:
            next(self.get_automorphism_mapping())
        except StopIteration:
            return False
        return True

    def get_automorphism_mapping(self: 'Graph') -> Iterator[Dict[int, int]]:
        """
        Iterator of all possible automorphism mappings.
        """
        return _get_automorphism_mapping(self.atoms_order, self._bonds,
                                         {n: atom_frequency(a) for n, a in self._atoms.items()})


def _get_automorphism_mapping(atoms: Dict[int, int], bonds: Dict[int, Dict[int, Any]],
                              atoms_frequencies: Dict[int, int]) -> Iterator[Dict[int, int]]:
    if len(atoms) == len(set(atoms.values())):
        return  # all atoms unique

    components, closures = _compile_query(atoms, bonds, atoms_frequencies)
    mappers = [_get_mapping(order, closures, atoms, bonds, {x for x, *_ in order})
               for order in components]
    if len(mappers) == 1:
        for mapping in mappers[0]:
            if any(k != v for k, v in mapping.items()):
                yield mapping
    for match in lazy_product(*mappers):
        mapping = match[0].copy()
        for m in match[1:]:
            mapping.update(m)
        if any(k != v for k, v in mapping.items()):
            yield mapping


def _get_mapping(linear_query, query_closures, o_atoms, o_bonds, scope):
    size = len(linear_query) - 1
    order_depth = {v[0]: k for k, v in enumerate(linear_query)}

    stack = deque()
    path = []
    mapping = {}
    reversed_mapping = {}

    s_n, _, s_atom, _ = linear_query[0]
    for n, o_atom in o_atoms.items():
        if n in scope and s_atom == o_atom:
            stack.append((n, 0))

    while stack:
        n, depth = stack.pop()
        current = linear_query[depth][0]
        if depth == size:
            yield {**mapping, current: n}
        else:
            if len(path) != depth:
                for x in path[depth:]:
                    del mapping[reversed_mapping.pop(x)]
                path = path[:depth]

            path.append(n)
            mapping[current] = n
            reversed_mapping[n] = current

            depth += 1
            s_n, back, s_atom, s_bond = linear_query[depth]
            if back != current:
                n = path[order_depth[back]]

            for o_n, o_bond in o_bonds[n].items():
                if o_n in scope and o_n not in reversed_mapping and s_bond == o_bond:
                    if s_atom == o_atoms[o_n]:
                        # check closures equality
                        o_closures = o_bonds[o_n].keys() & reversed_mapping.keys()
                        o_closures.discard(n)
                        if o_closures == {mapping[m] for m, _ in query_closures[s_n]}:
                            obon = o_bonds[o_n]
                            if all(bond == obon[mapping[m]] for m, bond in query_closures[s_n]):
                                stack.append((o_n, depth))


def _compile_query(atoms, bonds, atoms_frequencies):
    closures = defaultdict(list)
    components = []
    seen = set()
    while len(seen) < len(atoms):
        start = min(atoms.keys() - seen, key=lambda x: atoms_frequencies[x])
        seen.add(start)
        stack = [(n, start, atoms[n], bond) for n, bond in sorted(bonds[start].items(), reverse=True,
                                                                  key=lambda x: atoms_frequencies[x[0]])]
        order = [(start, None, atoms[start], None)]
        components.append(order)

        while stack:
            front, back, *_ = atom = stack.pop()
            if front not in seen:
                order.append(atom)
                for n, bond in sorted(bonds[front].items(), reverse=True, key=lambda x: atoms_frequencies[x[0]]):
                    if n != back:
                        if n not in seen:
                            stack.append((n, front, atoms[n], bond))
                        else:
                            closures[front].append((n, bond))
                seen.add(front)
    return components, closures


__all__ = ['Isomorphism']



# -*- coding: utf-8 -*-
#
#  Copyright 2020, 2021 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from itertools import product


# lazy itertools.product with diagonal combination precedence
def lazy_product(*args):
    if len(args) == 1:
        for x in args[0]:
            yield x,
    elif not args:
        yield ()
    else:
        gens = [iter(x) for x in args]
        empty = [False] * len(args)
        pools = [[] for _ in range(len(args))]
        indices = set()

        reached = 0
        while True:
            out = []
            ind = []
            for n, (p, g, e) in enumerate(zip(pools, gens, empty)):
                if e:
                    out.append(p[-1])
                else:
                    try:
                        x = next(g)
                    except StopIteration:
                        if not p:  # one of gens empty
                            return
                        reached += 1
                        if reached == len(args):
                            break
                        out.append(p[-1])
                        empty[n] = True
                    else:
                        p.append(x)
                        out.append(x)
                ind.append(len(p) - 1)
            else:
                yield tuple(out)
                indices.add(tuple(ind))
                continue
            break

        for ind in product(*(range(len(p)) for p in pools)):
            if ind in indices:
                continue
            yield tuple(p[x] for x, p in zip(ind, pools))


__all__ = ['lazy_product']



# -*- coding: utf-8 -*-
#
#  Copyright 2019-2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  Copyright 2019 Adelia Fatykhova <adelik21979@gmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from collections import deque
from functools import reduce
from itertools import count, permutations, combinations
from logging import getLogger, INFO
from operator import or_
from typing import List, Iterator, Tuple, Union
from .base import BaseReactor
from .._functions import lazy_product
from ..containers import QueryContainer, MoleculeContainer, ReactionContainer


logger = getLogger('chython.reactor')
logger.setLevel(INFO)


class Reactor(BaseReactor):
    """
    Reactor for molecules transformations.
    Generates reaction from input molecules using transformation template.

    Reactor calling transforms reactants to products and
    returns generator of reaction transformations with all
    possible reactions.
    """
    def __init__(self, patterns: Tuple[QueryContainer, ...],
                 products: Tuple[Union[MoleculeContainer, QueryContainer], ...], *,
                 delete_atoms: bool = True, one_shot: bool = True, polymerise_limit: int = 10,
                 automorphism_filter: bool = True, fix_aromatic_rings: bool = True, fix_tautomers: bool = True):
        """
        :param patterns: Search patterns for each reactant.
        :param products: Resulted structures.
        :param delete_atoms: If True atoms exists in reactants but not exists in products will be removed.
        :param one_shot: Do only single reaction center then True, else do all possible combinations of reactions.
        :param polymerise_limit: Limit of self reactions. Make sense than one_shot = False.
        :param fix_aromatic_rings: Proceed kekule and thiele on products.
        :param fix_tautomers: See `thiele()` docs.
        :param automorphism_filter: Skip matches to same atoms.
        """
        if not patterns or not products:
            raise ValueError('empty template')

        if not all(isinstance(x, QueryContainer) for x in patterns):
            raise TypeError('invalid params')
        elif not all(isinstance(x, (QueryContainer, MoleculeContainer)) for x in products):
            raise TypeError('invalid params')
        self.patterns = patterns
        self.products = products

        self.__one_shot = one_shot
        self.__polymerise_limit = polymerise_limit
        self.__products_atoms = tuple(set(m) for m in products)
        self.__automorphism_filter = automorphism_filter
        super().__init__({n for x in patterns for n, h in x._masked.items() if not h}, reduce(or_, products),
                         delete_atoms, fix_aromatic_rings, fix_tautomers)

    def __call__(self, *structures: MoleculeContainer):
        if any(not isinstance(structure, MoleculeContainer) for structure in structures):
            raise TypeError('only list of Molecules possible')

        len_patterns = len(self.patterns)
        structures = fix_mapping_overlap(structures)
        s_nums = set(range(len(structures)))
        seen = set()
        if self.__one_shot:
            for chosen in permutations(s_nums, len_patterns):
                ignored = [structures[x] for x in s_nums.difference(chosen)]
                chosen = [structures[x] for x in chosen]
                for new in self.__single_stage(chosen, {x for x in ignored for x in x}):
                    # store reacted molecules in same order as matched pattern
                    r = ReactionContainer([x.copy() for x in chosen] + [x.copy() for x in ignored],
                                          new + [x.copy() for x in ignored])
                    if len(new) > 1:  # try to keep salts
                        r.contract_ions()
                    if str(r) in seen:
                        continue
                    seen.add(str(r))
                    yield r
        else:
            queue = deque(([structures[x] for x in chosen], [structures[x] for x in s_nums.difference(chosen)], 0)
                          for chosen in permutations(s_nums, len_patterns))
            while queue:
                chosen, ignored, depth = queue.popleft()
                depth += 1
                for new in self.__single_stage(chosen, {x for x in ignored for x in x}):
                    r = ReactionContainer([x.copy() for x in structures], new + [x.copy() for x in ignored])
                    if len(new) > 1:
                        r.contract_ions()  # try to keep salts
                        if str(r) in seen:
                            continue
                        seen.add(str(r))
                        if len(r.products) != len(ignored) + len(self.__products_atoms):
                            logger.info('ambiguous multicomponent structures. skip multistage processing')
                            yield r
                            continue
                    elif str(r) in seen:
                        continue
                    else:
                        seen.add(str(r))

                    if depth < self.__polymerise_limit:
                        prod = r.products
                        if len_patterns == 1:  # simple case. only products or ignored can be transformed.
                            for i in range(len(prod)):
                                queue.append(([prod[i]], [*prod[:i], *prod[i + 1:]], depth))
                        else:  # one of products molecule combined with previously chosen
                            for chp in combinations(chosen, len_patterns - 1):
                                for i in range(len(prod)):
                                    for ch in permutations(fix_mapping_overlap((prod[i], *chp)), len_patterns):
                                        queue.append((ch, [*prod[:i], *prod[i + 1:]], depth))
                    yield r

    def __single_stage(self, chosen, ignored) -> Iterator[List[MoleculeContainer]]:
        max_ignored_number = max(ignored, default=0)
        united_chosen = reduce(or_, chosen)
        split = len(self.__products_atoms) > 1
        for match in lazy_product(*(x.get_mapping(y, automorphism_filter=self.__automorphism_filter) for x, y in
                                    zip(self.patterns, chosen))):
            mapping = match[0].copy()
            for m in match[1:]:
                mapping.update(m)
            for new in self._patcher(united_chosen, mapping):
                collision = set(new).intersection(ignored)
                if collision:
                    new.remap(dict(zip(collision, count(max(max_ignored_number, max(new)) + 1))))

                if split:
                    yield new.split()
                else:
                    yield [new]


def fix_mapping_overlap(structures) -> List[MoleculeContainer]:
    checked = []
    checked_atoms = set()
    for structure in structures:
        intersection = set(structure).intersection(checked_atoms)
        if intersection:
            mapping = dict(zip(intersection, count(max(max(checked_atoms), max(structure)) + 1)))
            structure = structure.remap(mapping, copy=True)
            logger.info('some atoms in input structures had the same numbers.\n'
                        f'atoms {list(mapping)} were remapped to {list(mapping.values())}')
        checked_atoms.update(structure)
        checked.append(structure)
    return checked


__all__ = ['Reactor', 'fix_mapping_overlap']



# -*- coding: utf-8 -*-
#
#  Copyright 2021, 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from base64 import urlsafe_b64encode
from fileinput import FileInput
from io import StringIO, TextIOWrapper
from itertools import islice
from os.path import abspath, join
from pathlib import Path
from pickle import dump, load, UnpicklingError
from sys import platform
from tempfile import gettempdir
from typing import Union, Iterator, List
from .stereo import MDLStereo
from ...containers import ReactionContainer, MoleculeContainer
from ...exceptions import ParseError


class MDLReadMeta(type):
    def __call__(cls, *args, **kwargs):
        if kwargs.get('indexable'):
            _cls = type(cls.__name__, (cls,), {'__len__': lambda x: len(x._shifts) - 1, '__module__': cls.__module__})
            obj = object.__new__(_cls)
        else:
            obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj


class MDLRead(MDLStereo, metaclass=MDLReadMeta):
    def __init__(self, file, **kwargs):
        if isinstance(file, str):
            self._file = open(file)
            self._is_buffer = False
        elif isinstance(file, Path):
            self._file = file.open()
            self._is_buffer = False
        elif isinstance(file, (TextIOWrapper, StringIO, FileInput)):
            self._file = file
            self._is_buffer = True
        else:
            raise TypeError('invalid file. TextIOWrapper, StringIO subclasses possible')
        super().__init__(**kwargs)

    def close(self, force=False):
        """
        Close opened file

        :param force: force closing of externally opened file or buffer
        """
        if not self._is_buffer or force:
            self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def _load_cache(self):
        """
        Load existing cache or create new. Working only for UNIX-like systems and local files (not buffers).
        """
        if platform == 'win32' or self._is_buffer:
            return
        try:
            with open(self.__cache_path, 'rb') as f:
                self._shifts = load(f)
        except FileNotFoundError:  # cache not found
            self.reset_index()
        except IsADirectoryError as e:
            raise IsADirectoryError(f'Please delete {self.__cache_path} directory') from e
        except (UnpicklingError, EOFError) as e:  # invalid file. ask user to check it.
            raise UnpicklingError(f'Invalid cache file {self.__cache_path}. Please delete it') from e

    def reset_index(self):
        """
        Create (rewrite) indexation table. Implemented only for object that
        is a real file (the path to the file is specified) because the external grep utility is used.
        """
        if platform != 'win32' and not self._is_buffer:
            self._shifts = self._get_shifts(self._file.name)
            with open(self.__cache_path, 'wb') as f:
                dump(self._shifts, f)
        else:
            raise self._implement_error

    @property
    def __cache_path(self):
        return abspath(join(gettempdir(), 'chython_' + urlsafe_b64encode(abspath(self._file.name).encode()).decode()))

    def read(self) -> List[Union[ReactionContainer, MoleculeContainer]]:
        """
        Parse whole file

        :return: list of parsed molecules
        """
        return list(iter(self))

    def __iter__(self) -> Iterator[Union[ReactionContainer, MoleculeContainer]]:
        return (x for x in self._data if not isinstance(x, ParseError))

    def __next__(self) -> Union[ReactionContainer, MoleculeContainer]:
        return next(iter(self))

    def __getitem__(self, item) -> Union[ReactionContainer, MoleculeContainer, ParseError]:
        """
        Getting the item by index from the original file,
        For slices records with errors skipped.
        For indexed access records with errors returned as error container.
        :return: [Molecule, Reaction]Container or list of [Molecule, Reaction]Containers
        """
        if self._shifts:
            _len = len(self._shifts) - 1
            if isinstance(item, int):
                if item >= _len or item < -_len:
                    raise IndexError('List index out of range')
                if item < 0:
                    item += _len
                self.seek(item)
                return next(self._data)
            elif isinstance(item, slice):
                start, stop, step = item.indices(_len)
                if start == stop:
                    return []
                if step == 1:
                    self.seek(start)
                    records = [x for x in islice(self._data, stop - start) if not isinstance(x, ParseError)]
                else:
                    records = []
                    for index in range(start, stop, step):
                        self.seek(index)
                        record = next(self._data)
                        if not isinstance(record, ParseError):
                            records.append(record)
                return records
            else:
                raise TypeError('Indices must be integers or slices')
        raise self._implement_error

    def read_text(self, item):
        """
        Read record block as text
        """
        if self._shifts:
            if not isinstance(item, int):
                raise TypeError('int required')
            _len = len(self._shifts) - 1
            if item >= _len or item < -_len:
                raise IndexError('List index out of range')
            if item < 0:
                item += _len
            start = self._shifts[item]
            end = self._shifts[item + 1]
            current = self._file.tell()
            self._file.seek(start)
            data = self._file.read(end - start)
            self._file.seek(current)
            return data
        raise self._implement_error

    def _prepare_meta(self, meta):
        new_meta = {}
        for k, v in meta.items():
            if v:
                new_meta[k] = '\n'.join(v)
            else:
                self._info(f'invalid metadata entry: {k}: {v}')
        return new_meta

    _shifts = None
    _implement_error = NotImplementedError('Indexable supported in unix-like o.s. and for files stored on disk')


__all__ = ['MDLRead']



# -*- coding: utf-8 -*-
#
#  Copyright 2022 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of chython.
#
#  chython is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#
from lazy_object_proxy import Proxy


def _rules():
    from ... import smiles, smarts
    raw_rules = []

    # phenol etherification
    r = smiles('[OH:5][cH:1]:[cH2:2].[CH3:4][OH:3]>>[CH3:4][O:3][cH:1]:[cH2:2]')
    m = smarts('[O;D1:5].[O;D1:3]')  # reactant restriction.
    raw_rules.append((r, {3: 5}, m))

    # esterification
    r = smiles('[OH:2][CH:1]=[O:3].[CH3:4][OH:5]>>[CH3:4][O:2][CH:1]=[O:3]')
    m = smarts('[O;D1:2].[O;D1:5]')
    raw_rules.append((r, {2: 5}, m))

    # acid reduction
    r = smiles('[CH3:2][C:1]([OH:3])=[O:4]>>[CH3:2][CH2:1][OH:3]')
    m = smarts('[O;D1:3]')
    raw_rules.append((r, {3: 4, 4: 3}, m))

    # ester reduction
    r = smiles('[CH3:5][O:3][C:1]([CH3:2])=[O:4]>>[CH3:2][CH2:1][OH:3]')
    raw_rules.append((r, {3: 4}, None))

    # ozonolysis
    r = smiles('[O-:4][O+:1]=[O:6].[CH3:3][CH:2]=[CH2:5]>>[CH3:3][CH:2]=[O:1]')
    raw_rules.append((r, {1: 4}, None))

    # peroxide oxirane
    r = smiles('[CH3:6][O:1][OH:5].[CH3:4][CH:2]=[CH2:3]>>[CH3:4][CH:2]1[CH2:3][O:1]1')
    m = smarts('[O;D1:5]')
    raw_rules.append((r, {1: 5}, m))

    # claisen rearrangement
    r = smiles('[cH2:4]:[cH:3]:[cH:2][O:1][CH2:5][CH:6]=[CH2:7]>>[OH:1][cH:2]:[c:3](:[cH2:4])[CH2:5][CH:6]=[CH2:7]')
    raw_rules.append((r, {5: 7, 7: 5}, None))

    # metathesis
    r = smiles('[CH3:1][CH:2]=[CH2:3].[CH3:5][CH:4]=[CH2:6]>>[CH3:5][CH:3]=[CH:2][CH3:1]')
    raw_rules.append((r, {3: 4, 4: 3}, None))

    rules = []
    for r, f, m in raw_rules:
        c = ~r
        rules.append((c, str(c.substructure(c.center_atoms)), m, f))
    return rules


rules = Proxy(_rules)


__all__ = ['rules']
